from MJOLNIR.Geometry.Analyser import Analyser, FlatAnalyser
import numpy as np
import warnings
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

def test_init():
    GenericAnalyser = Analyser(position=(0.0,1.0,0.0),direction=(1.0,0,0),d_spacing=3.35,mosaicity=60.0)
    assert(np.all(GenericAnalyser.position==np.array([0.0,1.0,0.0])))
    assert(np.all(GenericAnalyser.direction==(1.0,0.0,0.0)))
    assert(GenericAnalyser.d_spacing==3.35)
    assert(GenericAnalyser.mosaicity==60.0)

def test_Generic_plot():
    
    GenericAnalyser = Analyser(position=(0.0,1.0,0.0),direction=(1.0,0,0),d_spacing=3.35,mosaicity=60.0)
    ax = []
    try:
        GenericAnalyser.plot(ax)
        assert False
    except NotImplementedError:
        assert True

def test_warnings():
    GenericAnalyser = Analyser(position=(0.0,1.0,0.0),direction=(1.0,0,0),d_spacing=3.35,mosaicity=60.0)

    with warnings.catch_warnings(record=True) as w: # From https://docs.python.org/3.1/library/warnings.html
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        # Trigger a warning.
        GenericAnalyser.d_spacing = 20.0
        GenericAnalyser.mosaicity = 0.5
        # Verify some things
        assert len(w) == 2
        assert issubclass(w[0].category, UserWarning)
        assert issubclass(w[1].category, UserWarning)
        assert "The unit of d spacing is Angstrom" in str(w[0].message)
        assert "The unit of mosaicity is arcminutes" in str(w[1].message)

def test_Generic_errors():
    GenericAnalyser = Analyser(position=(0.0,1.0,0.0),direction=(1.0,0,0),d_spacing=3.35,mosaicity=60.0)
    try:
        GenericAnalyser.d_spacing=-0.1
        assert False
    except AttributeError:
        assert True

    try:
        GenericAnalyser.mosaicity=-0.1
        assert False
    except AttributeError:
        assert True


def test_Analyser_init():
    Analyser = FlatAnalyser(position=(0.0,1.0,0.0),direction=(1.0,0,0),d_spacing=3.35,mosaicity=60,width=0.05,height=0.1)
    assert(np.all(Analyser.position==np.array([0.0,1.0,0.0])))
    assert(np.all(Analyser.direction==(1.0,0.0,0.0)))
    assert(Analyser.width==0.05)
    assert(Analyser.height==0.1)
 

def test_Analyser_width():
    Analyser = FlatAnalyser(position=(0.0,1.0,0.0),direction=(1.0,0,0))
    try:
        Analyser.width=-0.1
        assert False
    except AttributeError:
        assert True

def test_Analyser_height():
    Analyser = FlatAnalyser(position=(0.0,1.0,0.0),direction=(1.0,0,0))
    try:
        Analyser.height=-0.1
        assert False
    except AttributeError:
        assert True


def test_FlatAnalyser_plot():
    Analyser = FlatAnalyser(position=(0.0,1.0,0.0),direction=(1.0,0,0))
    fig = plt.figure()
    ax = fig.add_subplot(projection='3d')

    Analyser.plot(ax)
    
def test_FlatAnalyser_str():
    Analyser = FlatAnalyser(position=(0.0,1.0,0.0),direction=(1.0,0,0))
    assert(str(Analyser)=='FlatAnalyser located at '+np.array2string(np.array([0.0,1.0,0.0])))


def test_FlatAnalyser_kwChecker():

    try:
        Analyser = FlatAnalyser(Position=(0.0,1.0,0.0),direction=(1.0,0,0))
        assert False
    except:
        assert True

    try:
        Analyser = FlatAnalyser(pos=(0.0,1.0,0.0),direction=(1.0,0,0))
        assert False
    except:
        assert True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on Wed Jul 14 11:54:33 2021

@author: Jakob Lass

Tool to replace makefile in order to fully port development to windows
"""

import argparse
from ntpath import join

import os,glob
import re, numpy as np
from MJOLNIR.Data.DataSet import OxfordList

DEBUG = False

possibleArguments = ['test', 'tutorials', 'wheel', 'html', 'upload', 'testVersion', 'version']

parser = argparse.ArgumentParser(description="Make tool replacing linux Makefile.")
parser.add_argument("task", nargs='?', default='help', type=str, help="Type of task to be performed. Possible tasks are: {}. Run without argument to see help menu.".format(OxfordList(possibleArguments)))
parser.add_argument("version", nargs='?',default=None, type=str, help="Version number to be created. Default is previous plus 1, i.e. x.y.z -> x.y.z+1")

def callHelp(parser=parser):
    parser.print_help()

def test():
    os.system("python -m pytest -vv test")

# def cleanTutorialFolders():
#     folders = [f.path for f in os.scandir(os.path.join('docs','Tutorials')) if f.is_dir()]
#     print('Cleaning tutorial folders: '+OxfordList(folders))
#     for folder in folders:
#         os.system('python clean.py '+folder)

def generateTutorials():
    if not DEBUG:
        os.system('python '+str(os.path.join('docs','tutorials.py')))
    else:
        print('python '+str(os.path.join('docs','tutorials.py')))

def makeHTML():
    print('Creating HTML through Sphinx')
    if not DEBUG:
        os.system("sphinx-build docs build")

def makeWheel():
    if not DEBUG:
        os.system("python -m build")
    else:
        print("python -m build")

def getLatestVersions():
    list_of_files = glob.glob(os.path.join('dist','*')) # * means all if need specific format then *.csv
    list_of_files = sorted(list_of_files, key=os.path.getctime)
    
    versions = {}
    pattern = r"-([0-9]*)\.([0-9]*)\.([0-9]*)\.?(post)?([0-9]*)?"

    for b in list_of_files:
        #b = 'dist\\MJOLNIR-1.1.19.tar.gz'
        a = re.findall(pattern,b.lower())[0]
        if not a[0] == '1':
            print(b,a)
        else:
            a = np.asarray(a)
            a = a[np.asarray([len(x)>0 for x in a])]
            a = '.'.join(a)
            a = a.replace('post.','post')
            if not a in versions:
                versions[a] = [os.path.getctime(os.path.join(r'C:\Users\lass_j\Documents\Software\MJOLNIR',b)),[b]]
            else:
                versions[a][1].append(b)
    sortedVersions = dict(sorted(versions.items(), key=lambda item: -item[1][0]))
    return sortedVersions

def getLatestBuilds():
    sortedVersions = getLatestVersions()
    key  = list(sortedVersions.keys())[0]
    return sortedVersions[key][1]

def getCurrentVersion():
    sortedVersions = getLatestVersions()
    string  = list(sortedVersions.keys())[0]
    #versionList = string.split('-')[1].split('.')[:-2]
    # if 'post' in versionList[-1]:
    #     versionList[-1] = versionList[-1].replace('post','')
    #     adder = 'post'
    # else:
    #     adder = ''
    # version = [int(x) for x in versionList]
    
    # if adder != '':
    #     version[-1]=adder+str(version[-1]+1)
    # else:
    #     version[-1]+=1
    return string#'.'.join([str(x) for x in version])

def uploadPyPi(server='testpypi'):
    build = getLatestBuilds()
    for b in build:
        print('Uploading ' + str(b) + ' to '+server)
        if not DEBUG:
            os.system('twine upload {} -r {}'.format(b,server))


def upload():
    uploadPyPi(server='testpypi')
    uploadPyPi(server='pypi')

def update(version):
    if not DEBUG:
        os.system('python Update.py '+version)
    else:
        print('python Update.py '+version)

def makeTutorials():
    # cleanTutorialFolders()
    generateTutorials()
    makeHTML()

args = parser.parse_args()


if args.task == '' or args.task.lower() == 'help':
    callHelp()
elif args.task.lower() == 'test':
    test()

elif args.task.lower() == 'tutorials':
    makeTutorials()

elif args.task.lower() == 'wheel':
    makeWheel()

elif args.task.lower() == 'upload':
    upload()

elif args.task.lower() == 'testversion':
    if args.version is None:
        version = getCurrentVersion()
    else:
        version = args.version
    print('Creating test version '+version)
    update(version=version)
    makeWheel()
    uploadPyPi(server='testpypi')

elif args.task.lower() == 'html':
    makeHTML()

elif args.task.lower() == 'version':
    if args.version is None:
        version = getCurrentVersion()
    else:
        version = args.version
    print('Creating version '+version)
    update(version=version)
    makeTutorials()
    addFiles = ['setup.py',
                os.path.join('docs','conf.py'),
                os.path.join('docs','Tutorials','*'),
                os.path.join('docs','index.rst'),
                os.path.join('MJOLNIR','__init__.py'),
                ]
    if not DEBUG:
        os.system("git add {}".format(' '.join(addFiles)))
        os.system('git commit -m "Update version"')
        os.system('git tag -a {0} -m "{0}"'.format(version))
    else:
        print("git add {}".format(' '.join(addFiles)))
        print('git commit -m "Update version"')
        print('git tag -a {0} -m "{0}"'.format(version))
    makeWheel()
    if not DEBUG:
        os.system("git push")
        os.system("git push --tags")
    else:
        print("git push")
        print("git push --tags")

else:
    print('Provided argument not understood. Received',args.task,'\n\n')
    callHelp()


import numpy as np
from MJOLNIR.Data.DataFile import DataFile,decodeStr,createEmptyDataFile,assertFile, possibleAttributes, shallowRead
from MJOLNIR import _tools
import MJOLNIR.Data.Sample
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
import os

dataPath = 'samlpedata'


def test_DataFile():
    try:
        DF = DataFile('/nope.txt')
        assert False
    except:
        assert True

    try:
        DF= DataFile(os.path.join(dataPath,'CAMEA_Full.xml')) # Wrong file
        assert False
    except:
        assert True

    files = [os.path.join(dataPath,f) for f in ['camea2018n000137.hdf','camea2018n000137.nxs','camea2022n000894.hdf','camea2022n000894.nxs']]
    DF1 = DataFile(files[0])

    size = DF1.size
    shape = DF1.shape
    assert(size==DF1.I.size)
    assert(np.all(shape == DF1.I.shape))

    try:
        DF1.size = 0
        assert False
    except AttributeError:
        assert True
    try:
        DF1.shape = 0
        assert False
    except AttributeError:
        assert True

    assertFile(files[1])
    DF2 = DataFile(files[1])
    s = str(DF2)
    sampleS = str(DF2.sample)
    print(str(DF1.sample))
    print(str(DF2.sample))
    assert(DF1.sample == DF2.sample)

def test_DataFile_equility():
    f1 = DataFile(os.path.join(dataPath,'camea2018n000136.hdf'))
    print('----------')
    f2 = DataFile(os.path.join(dataPath,'camea2018n000136.hdf'))
    assert(f1==f2)
    print('----------')
    f3 = DataFile(f2)
    assert(f1==f3)
    print('----------')
    f4 = DataFile(os.path.join(dataPath,'camea2018n000137.hdf'))
    assert(f1!=f4)
    f5 = f2.convert(binning=8)
    f5.saveNXsqom(os.path.join(dataPath,'camea2018n000136.nxs'))
    f3 = DataFile(os.path.join(dataPath,'camea2018n000136.nxs'))
    assert(f1==f3)
    print('----------')
    f6 = DataFile(os.path.join(dataPath,'camea2022n000894.hdf'))
    f6b = f6.convert(binning=8)
    f6b.saveNXsqom(os.path.join(dataPath,'camea2022n000894.nxs'))
    f7 = DataFile(os.path.join(dataPath,'camea2022n000894.nxs'))
    assert(f6==f7)

def test_DataFile_plotA4():
    plt.ioff()
    import matplotlib
    #matplotlib.use('Agg')
    fileName = os.path.join(dataPath,'camea2022n000894.hdf')
    fileName2= os.path.join(dataPath,'camea2022n000894.nxs')
    file = DataFile(fileName)
    

    try:
        file.plotA4(binning=20) # Binning not found in data file
        assert False
    except AttributeError:
        assert True

    fig = file.plotA4(1)
    fig2 = file.plotA4()
    assertFile(fileName2)
    file2 = DataFile(fileName2)
    try:
        file2.plotA4(binning=20) # Binning not found in data file
        assert False
    except AttributeError:
        assert True
    
    file2.plotA4(binning=1)
    plt.close('all')

    
def test_DataFile_plotEf():
    plt.ioff()
    import matplotlib
    #matplotlib.use('Agg')
    fileName = os.path.join(dataPath,'camea2022n000894.hdf')
    fileName2= os.path.join(dataPath,'camea2022n000894.nxs')
    assertFile(fileName2)
    file = DataFile(fileName)

    try:
        file.plotEf(binning=20) # Binning not found in data file
        assert False
    except AttributeError:
        assert True

    fig = file.plotEf(1)
    fig2 = file.plotEf()
    
    file2 = DataFile(fileName2)
    try:
        file2.plotEf(binning=20) # Binning not found in data file
        assert False
    except AttributeError:
        assert True

    file2.plotEf(binning=1)
    plt.close('all')

def test_DataFile_plotEfOverview():
    plt.ioff()
    import matplotlib
    #matplotlib.use('Agg')
    fileName = os.path.join(dataPath,'camea2022n000894.hdf')
    fileName2= os.path.join(dataPath,'camea2022n000894.nxs')
    assertFile(fileName2)

    file = DataFile(fileName)

    try:
        file.plotEfOverview(binning=20) # Binning not found in data file
        assert False
    except AttributeError:
        assert True

    fig = file.plotEfOverview(1)
    fig2 = file.plotEfOverview()

    file2 = DataFile(fileName2)
    try:
        file2.plotEfOverview(binning=20) # Binning not found in data file
        assert False
    except AttributeError:
        assert True

    file2.plotEfOverview(binning=1)
    plt.close('all')

def test_DataFile_plotNormalization():
    plt.ioff()
    import matplotlib
    #matplotlib.use('Agg')
    fileName = os.path.join(dataPath,'camea2022n000894.hdf')
    fileName2= os.path.join(dataPath,'camea2022n000894.nxs')
    file = DataFile(fileName)
    assertFile(fileName2)

    try:
        file.plotNormalization(binning=20) # Binning not found in data file
        assert False
    except AttributeError:
        assert True

    fig = file.plotNormalization(1)
    fig2 = file.plotNormalization()

    file2 = DataFile(fileName2)
    try:
        file2.plotNormalization(binning=20) # Binning not found in data file
        assert False
    except AttributeError:
        assert True

    file2.plotNormalization(binning=1)
    plt.close('all')

def test_DataFile_decodeString():
    a = b'String'
    b = 'String'

    c =1.1 # Float

    assert(decodeStr(a)==decodeStr(b))
    assert(c == decodeStr(c))

def test_DataFile_ScanParameter():

    files = [os.path.join(dataPath,'camea2022n000894.hdf'),os.path.join(dataPath,'camea2022n000894.nxs')]
    assertFile(files[1])
    for file in files:
        dfile = DataFile(file)
        assert(dfile.scanParameters[0]=='A3')
        assert(len(dfile.scanParameters)==len(dfile.scanUnits))
        assert(len(dfile.scanParameters)==len(dfile.scanValues))
        assert(len(dfile.scanParameters)==1)
        assert(dfile.scanUnits[0]=='degree')
        ##assert(np.all(dfile.scanValues==np.arange(0,150,1)))


def test_DataFile_Error():
    df = DataFile(os.path.join(dataPath,'camea2022n000894.hdf'))

    # Not implimented
    try:
        df+df
        assert False
    except NotImplementedError:
        assert True

    df.instrument = 'WrongInstrument'
    try:
        df.convert(binning=1)
        assert False
    except AttributeError:
        assert True
    
    df2 = DataFile(os.path.join(dataPath,'camea2022n000894.nxs'))
    try:
        df.saveNXsqom(os.path.join(dataPath,'saving.nxs'))
        assert False
    except AttributeError: # File does not have original_file attribute
        assert True

    df2.type = 'WrongType'
    try:
        df2.saveNXsqom(os.path.join(dataPath,'saving.nxs'))
        assert False
    except AttributeError: # Manually change type to wrong
        assert True


def test_DataFile_SaveLoad():
    df = DataFile(os.path.join(dataPath,'camea2022n000894.hdf'))
    df.saveHDF(os.path.join(dataPath,'camea2022n000894_2.hdf'))
    df2= DataFile(os.path.join(dataPath,'camea2022n000894_2.hdf'))
    failed = []

    for att,val in df.__dict__.items():
        if att in ['name','fileLocation','fromNICOS']: # Name and location are to be different
            continue
        if isinstance(val,np.ndarray):
            if val.dtype == 'O':
                continue
            try:
                test = np.all(np.isclose(val,getattr(df2,att)))
            except:
                test = np.all(val==getattr(df2,att))
        else:
            test = val == getattr(df2,att)
        if not test:
            failed.append(att)
    print(failed)
    assert(len(failed)==0)
    os.remove(df2.fileLocation)


def test_DataFile_CreateEmpty(): # TODO: Make this test!!!
    nf = np.array([os.path.join('Data','Normalization_1.calib'),
    os.path.join('Data','Normalization_3.calib'),os.path.join('Data','Normalization_8.calib')])

    A3 = np.linspace(0,180,181)
    A3Position = 30.0
    A4 = -16
    Ei = 5.5
    Monitor = 1e5
    projectionVector1 = np.array([0,2,1,A3Position+30,A4,0.0,0.0,Ei,Ei])
    projectionVector2 = np.array([1,0,0,A3Position,A4,0.0,0.0,Ei,Ei])
    sample = MJOLNIR.Data.Sample.Sample(a=6.0,b=6.0,c=12.2,projectionVector2=projectionVector2,projectionVector1=projectionVector1,gamma=120.,beta=80.,alpha=90.)

    try:
        _ = createEmptyDataFile(A3=10,A4=10,Ei=10,sample=sample) # No change in any parameter
        assert False
    except AttributeError:
        assert True
    
    try:
        _ = createEmptyDataFile(A3=[10,11],A4=[10,11,12],Ei=10,sample=sample) # Two parameters change but not with the same shape
        assert False
    except AttributeError:
        assert True
    

    df = createEmptyDataFile(A3=A3,A4=A4,Ei=Ei,sample=sample,Monitor=Monitor,normalizationFiles = nf)
    
    # Check the contents of df
    assert(df.sample == sample)
    assert(len(df.possibleBinnings)==len(nf))
    #assert(False)


def test_updateCalibration():
    calibFiles = [os.path.join('Data','Normalization80_1.calib'),
                    os.path.join('Data','Normalization80_3.calib'),
                    os.path.join('Data','Normalization80_5.calib')]


    df = DataFile(os.path.join(dataPath,'camea2022n000894.hdf'))
    print(df.I)
    print('----------------------')
    df.loadBinning(1)

    binnings = df.possibleBinnings # is 1,3,8
    edges = df.instrumentCalibrationEdges

    df.updateCalibration(calibFiles)

    df.loadBinning(1)
    newBinnings = df.possibleBinnings # is 1,3,8,5
    newEdges = df.instrumentCalibrationEdges
    
    assert(not np.any(newEdges!=edges)) # Check if all elements are equal


    df.updateCalibration(calibFiles,overwrite=True)
    df.loadBinning(1)

    newEdges = df.instrumentCalibrationEdges
    
    assert(np.any(newEdges!=edges)) # Check if all elements are equal

#
#def test_DataFile_BoundaryCalculation(quick):
#    if quick==True:
#        binning = [1,3,8]
#    else:
#        binning = [1]
#    for B in binning:
#        print('Using binning {}'.format(B))
#        df = DataFile(os.path.join(dataPath,'camea2018n000017.hdf'))
#        converted = df.convert(binning=B)
#        EP,EBins = converted.calculateEdgePolygons()
#        areas = np.array([e.area for e in EP])
#        assert(np.all(areas>2.0)) # Value found by running algorithm
#        assert(len(EBins)==B*8+1)
        

def test_shallowRead():
    # read out all possible things
    parameters = possibleAttributes
    files = [os.path.join(dataPath,'camea2018n000136.hdf'),
             os.path.join(dataPath,'camea2018n000137.hdf'),
             os.path.join(dataPath,'camea2022n000894.hdf')]
    result = shallowRead(files,parameters)

    assert(len(result) == len(files))
    assert(len(result[0]) == len(parameters))
    assert(len(result[0]) == len(result[1]))

    try:
        shallowRead(files,['NotExisting!'])
        assert False
    except AttributeError:
        assert True


import numpy as np

from MJOLNIR.Geometry.GeometryConcept import  GeometryConcept, GeometryObject
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt

def test_Concept_init():
    Concept = GeometryConcept(position=(0.0,1.0,0.0))
    assert(np.all(Concept.position==np.array([0.0,1.0,0.0])))


def test_Concept_plot():
    Concept = GeometryConcept()
    plt.ioff()
    fig = plt.figure()
    ax = fig.add_subplot(projection='3d')
    try:
        Concept.plot(ax)
        assert False
    except NotImplementedError:
        assert True
    print(str(Concept))



# Test of GeometryObject
def test_Object_init():
    GenericObject = GeometryObject(position=(0.0,1.0,0.0),direction=(1.0,0,0))
    assert(np.all(GenericObject.position==np.array([0.0,1.0,0.0])))
    assert(np.all(GenericObject.direction==(1.0,0.0,0.0)))

def test_position():
    GenericObject = GeometryObject(position=(0,1.0,0.0),direction=(1.0,0,0))
    GenericObject.position = (0.0,0.0,0.0)
    print(str(GenericObject))
    assert(np.all(GenericObject.position==(0.0,0.0,0.0)))

def test_Object_position_exception():
    GenericConcept = GeometryConcept(position=(0,1.0,0.0))
    try:
        GenericConcept.position=((0,0),(0,0))
        assert False
    except AttributeError:
        assert True

    try:
        GenericConcept.position=(0,0,0,0)
        assert False
    except AttributeError:
        assert True

def test_Object_direction():
    GenericObject = GeometryObject(position=(0,1.0,0.0),direction=(1.0,0,0))
    GenericObject.direction = (0.0,0.0,0.5)
    assert(np.all(GenericObject.direction==(0.0,0.0,1.0)))

def test_Object_direction_exception():
    GenericObject = GeometryObject(position=(0,1.0,0.0),direction=(1.0,0,0))
    try:
        GenericObject.direction=((0,0),(0,0))
        assert False
    except AttributeError:
        assert True

    try:
        GenericObject.direction=(0,0,0)
        assert False
    except AttributeError:
        assert True

    try:
        GenericObject.direction=(0,0,0,1)
        assert False
    except AttributeError:
        assert True


from importlib.util import decode_source
import numpy as np
import MJOLNIR.Data.DataFile
from MJOLNIR.Data.DataSet import DataSet,calculateGrid3D,binData3D,cut1DE,fmt,figureRowColumns,centeroidnp,compareNones,OxfordList, load
from MJOLNIR import _tools
import MJOLNIR.Data.Sample
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
import os,sys
import warnings
import pytest
from MJOLNIR.Data import Mask

pythonVersion = sys.version_info[0]

dataPath = 'samlpedata'

def test_DataSet_Creation():

    try:
        dataset = DataSet(OtherSetting=10.0,YetAnotherWrongSetting=20.0)
        assert False
    except AttributeError:
        assert True
    dataset = DataSet(Author='Jakob Lass')
    if(dataset.settings['Author']!='Jakob Lass'):
        assert False


def test_Dataset_Initialization():

    emptyDataset = DataSet()
    del emptyDataset
    MJOLNIR.Data.DataFile.assertFile(os.path.join(dataPath,'camea2018n000136.nxs'))
    dataset = DataSet(dataFiles=[os.path.join(dataPath,'camea2018n000136.hdf')],convertedFiles=os.path.join(dataPath,'camea2018n000137.nxs'),calibrationfiles=[])
    
    assert(dataset.dataFiles[0].name=='camea2018n000136.hdf')
    assert(dataset.convertedFiles[0].name=='camea2018n000137.nxs')
    assert(dataset.normalizationfiles == [])
    Str = str(dataset)

                                                                                                                 
def test_DataSet_Error():
    
    try:
        ds = DataSet(normalizationfiles=[10,11])
        assert False
    except:
        assert True

    ds = DataSet()
    
    try: # No data files
        ds.convertDataFile()
        assert False
    except AttributeError:
        assert True

    try: # No data files
        ds.binData3D(0.1,0.1,0.1)
        assert False
    except AttributeError:
        assert True

    try: # No data files
        ds.cut1D([0,0],[1,1],0.1,0.01,5.5,6.0)
        assert False
    except AttributeError:
        assert True

    try: # No data files
        ds.plotCut1D([0,0],[1,1],0.1,0.01,5.5,6.0)
        assert False
    except AttributeError:
        assert True

    try: # No data files
        ds.cutQE([0,0],[1,1],0.1,0.01,5.5,6.0)
        assert False
    except AttributeError:
        assert True

    try: # No data files
        ds.cutPowder(np.linspace(0,4,5),np.linspace(0,4,5))
        assert False
    except AttributeError:
        assert True

    try: # No data files
        ds.plotCutPowder(np.linspace(0,4,5),np.linspace(0,4,5))
        assert False
    except AttributeError:
        assert True
           
    

    try: # Wrong data file type
        ds.dataFiles = 100
        assert False
    except AttributeError:
        assert True


    try: # Can't overwrite settings
        ds.settings={}
        assert False
    except NotImplementedError:
        assert True

    try:# Wrong data file type
        ds.convertedFiles = 10
        assert False
    except AttributeError:
        assert True


    ds.dataFiles = os.path.join(dataPath,'camea2018n000136.hdf')

def test_LoadBambusData():
    ds = DataSet(dataFiles=[os.path.join(dataPath,'BambusTest.dat')])


    ## Set up values to check
    Ei = 3.684
    A4 = 38.194
    A3 = 20.0
    
    assert(ds[0].instrument == 'Bambus')
    assert(ds[0].binning == 1)
    assert(ds[0].dasel == (0,0))
    assert(ds.instrumentCalibrationEf.shape == [(100,4)])

    scanParameters = ds[0].scanParameters
    scanValues = ds[0].scanValues

    assert(len(scanParameters) == len(scanValues))
    

    for param,value in zip(['Ei','A4','A3'],[Ei,A4,A3]):
        assert(np.isclose(getattr(ds[0],param)[0],value))


    try: # Only binning 1 can be used
        ds.convertDataFile(binning = 3)
        assert False
    except AttributeError:
        assert True

    
    assert(ds[0].I.shape == (scanValues.shape[1],100,1))

    ds.convertDataFile()


def test_DataSet_Pythonic():
    dataFiles = [os.path.join(dataPath,'camea2018n000136.hdf'),os.path.join(dataPath,'camea2018n000137.hdf')]
    dataset = DataSet(dataFiles=dataFiles)
    assert(len(dataset)==2)
    for df in dataset:
        print(df)
    initShape = dataset.I.shape
    names = [dataset[i].name for i in range(len(dataset))]
    names.reverse()
    for i,df in enumerate(list(reversed(dataset))):
        names[i]==df.name

    dataset.append(dataFiles)
    assert(len(dataset)==4)
    secondShape = dataset.Monitor.shape
    assert(np.all(secondShape!=initShape))
    del dataset[3]
    del dataset[2]
    try:
        dataset[10]
        assert False
    except IndexError:
        assert True
    
    try:
        del dataset[10]
        assert False
    except IndexError:
        assert True

    try:
        dataset.append('NoFile')
    except:
        assert True
    
    dataset.append(MJOLNIR.Data.DataFile.DataFile(dataFiles[0]))
    assert(len(dataset)==3)
    assert(dataset.I.shape!=secondShape)



def test_DataSet_Equality():
    D1 = DataSet(dataFiles=os.path.join(dataPath,'camea2018n000136.hdf'))#,convertedFiles=['TestData/VanNormalization.nxs')])
    assert(D1==D1)

def test_DataSet_SaveLoad():
    
    D1 = DataSet(dataFiles=os.path.join(dataPath,'camea2018n000136.hdf'))#,convertedFiles = 'TestData/VanNormalization.nxs'))

    temp = 'temporary.bin'

    D1.save(temp)
    D2 = load(temp)
    os.remove(temp)
    assert(D1==D2) 

def test_DataSet_str():
    D1 = DataSet(dataFiles=os.path.join(dataPath,'camea2018n000136.hdf'))#,normalizationfiles = 'TestData/VanNormalization.hdf'))
    string = str(D1)
    print(string)


def test_DataSet_Convert_Data():
    dataFiles = os.path.join(dataPath,'camea2018n000136.hdf')
    dataset = DataSet(dataFiles=dataFiles)
    

    try:
        dataset.convertDataFile(dataFiles=dataFiles,binning=100)
        assert False
    except AttributeError: # Cant find normalization table
        assert True

    try:
        dataset.convertDataFile(dataFiles='FileDoesNotExist',binning=1)
        assert False
    except AttributeError: # FileDoesNotExist
        assert True

    try:
        os.remove(os.path.join(dataPath,'camea2018n000136.nxs'))
    except:
        pass
    dataset.convertDataFile(dataFiles=dataFiles,binning=8,saveLocation=os.path.join(dataPath,''),saveFile=True)
    convertedFile = dataset.convertedFiles[0]
    
    otherFile = MJOLNIR.Data.DataFile.DataFile(dataFiles.replace('.hdf','.nxs'))
    assert(otherFile.difference(convertedFile)==['counts'])
    #assert(otherFile==convertedFile)
    #assert(convertedFile==otherFile)
    os.remove(os.path.join(dataPath,'camea2018n000136.nxs'))
    


def test_DataSet_3DMesh():
    
    x = np.linspace(0,1,2)
    y = np.linspace(0,1,5)
    z = np.linspace(1,2,5)

    X,Y,Z = np.meshgrid(x,y,z,indexing='ij')
    XT1,YT1,ZT1 = calculateGrid3D(X,Y,Z)

    assert(XT1.shape==(3,6,6))
    assert(np.all(XT1[:,0,0]==np.array([-0.5,0.5,1.5])))
    assert(np.all(YT1[0,:,0]==np.array([-0.125,0.125,0.375,0.625,0.875,1.125])))
    assert(np.all(YT1[0,:,0]==ZT1[0,0,:]-1.0))



def test_DataSet_BinData():
    I = np.random.randint(0,100,(10,20,30))
    Norm = np.random.rand(10,20,30)
    Posx = np.linspace(0,1,10)
    Posy = np.linspace(0,1,20)
    Posz = np.linspace(1,2,30)
    PosX,PosY,PosZ = np.meshgrid(Posx,Posy,Posz,indexing='ij')



    pos = [PosX.flatten(),PosY.flatten(),PosZ.flatten()]
    Data,bins = binData3D(0.5,0.25,0.25,pos,I,norm=Norm)

    ReBinnedI = Data[0]
    RebinnedNorm = Data[1]
    RebinnedNormCount = Data[2]


    assert(ReBinnedI.shape==(3,5,5))
    #assert(np.all(bins[0].shape=(4,6,6)))
    assert(RebinnedNorm.shape==ReBinnedI.shape)
    assert(RebinnedNormCount.shape==ReBinnedI.shape)
    assert(RebinnedNormCount.dtype==float)
    assert(RebinnedNorm.dtype==Norm.dtype)
    assert(ReBinnedI.dtype==I.dtype)


def test_DataSet_full_test():
    import MJOLNIR.Data.Viewer3D
    
    import matplotlib.pyplot as plt
    import os
    plt.ioff()
    DataFile = [os.path.join(dataPath,'camea2018n000136.hdf')]

    dataset = DataSet(dataFiles=DataFile)
    dataset.convertDataFile(saveLocation=os.path.join(dataPath,''),saveFile=True)
    import matplotlib
    matplotlib.use('Agg')
    Data,bins = dataset.binData3D(0.08,0.08,0.25)
    
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        Intensity = np.divide(Data[0]*Data[3],Data[1]*Data[2])
    
    viewer = MJOLNIR.Data.Viewer3D.Viewer3D(Intensity,bins)
    viewer = dataset.View3D(0.08,0.08,0.25,CurratAxeBraggList=[[1,0,0]])
    
    if pythonVersion == 3: # Only possible in python 3
        viewer.ax.set_xticks_base(0.5)
        viewer.ax.set_yticks_base(0.5)

    viewer.setProjection(0)
    viewer.setPlane(4)
    del viewer 
    viewer = dataset.View3D(0.08,0.08,0.25,rlu=False,log=True)
    
    viewer.ax.get_figure().savefig('View3D.png')

    os.remove(os.path.join(dataPath,'camea2018n000136.nxs'))
    os.remove('View3D.png')
    del viewer
    plt.close('all')

def test_DataSet_Visualization():
    import warnings
    from MJOLNIR.Data import Viewer3D,DataFile
    DataFiles = [os.path.join(dataPath,'camea2018n000136.hdf')]

    dataset = DataSet(dataFiles=DataFiles)
    dataset.convertDataFile(saveLocation='Data\\',saveFile=True)

    Data,bins = dataset.binData3D(0.08,0.08,0.25)
    Data,bins = dataset.binData3D(0.08,0.08,0.25,dataFiles = [MJOLNIR.Data.DataFile.DataFile(os.path.join('Data','camea2018n000136.nxs'))])
    
    plt.ioff()
    import matplotlib
    matplotlib.use('Agg')

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        Intensity = np.divide(Data[0]*Data[3],Data[1]*Data[2])
    
    viewer = Viewer3D.Viewer3D(Intensity,bins)
    viewer.caxis = (0,100)
    try:
        viewer.caxis = 'Wrong type'
        assert False
    except AttributeError:
        assert True
    
    try:
        viewer.caxis = [0,1,2,3,4] # Too long input
        assert False
    except AttributeError:
        assert True
    
    try:
        viewer.setAxis(20) # Must bee 0,1, or 2
        assert False
    except AttributeError:
        assert True

    plt.plot()
    plt.close('all')
    os.remove(os.path.join('Data','camea2018n000136.nxs'))
    

def test_DataSet_binEdges():
    X = np.random.rand(100)*3 # array between 0 and 3 -ish
    X.sort()
    tolerance = 0.01
    Bins = _tools.binEdges(X,tolerance=tolerance)

    assert(Bins[0]==X[0]-0.1*tolerance)
    assert(np.isclose(Bins[-1],X[-1],atol=5) or Bins[-1]>X[-1])
    assert(len(Bins)<=3.0/tolerance)
    assert(np.all(np.diff(Bins[:-1])>tolerance*0.99))

def test_DataSet_1Dcut():
    q1 =  np.array([1.23,-1.51])
    q2 =  np.array([1.54, -1.25])
    width = 0.1

    plt.ioff()
    import matplotlib
    matplotlib.use('Agg')

    convertFiles = [os.path.join(dataPath,'camea2018n000136.hdf'),os.path.join(dataPath,'camea2018n000137.hdf')]
    
    ds = DataSet(dataFiles = convertFiles)
    ds.convertDataFile(saveFile=False)

    ax,Data,bins = ds.plotCut1D(q1,q2,width,rlu=False,minPixel=0.01,Emin=2.0,Emax=2.5,fmt='.')
    Data2,bins2 = ds.cut1D(q1,q2,width,rlu=False,minPixel=0.01,Emin=2.0,Emax=2.5)
    
    # Check that the two data sets have the same values (except for Data2 also having 'BinDistance')
    assert(Data2.equals(Data.loc[:, Data.columns != 'BinDistance']))

    Data,bins = ds.cut1D(q1,q2,width,rlu=False,minPixel=0.01,Emin=2.0,Emax=2.5,extend=False)
    assert(np.all(np.logical_and(bins[0][:,0]>=q1[0]-0.1,bins[0][:,0]<=q2[0]+0.1))) 
    # x-values should be between 1.1 and 2.0 correpsonding to q points given (add some extra space due to way bins are created (binEdges))

    #q3 = np.array([1.1,1.1])
    #q4 = np.array([2.0,2.0])
    Data,bins = ds.cut1D(q1,q2,width,rlu=False,minPixel=0.01,Emin=2.0,Emax=2.5,extend=False)
    
    assert(np.all(bins[0][:,0]>=q1[0]-0.1))
    assert(np.all(bins[0][:,0]<=q2[0]+0.1))
    assert(np.all(bins[0][:,1]>=q1[1]-0.1))
    assert(np.all(bins[0][:,1]<=q2[1]+0.1))
    # x and y-values should be between 1.1 and 2.0 correpsonding to q points given (add some extra space due to way bins are created (binEdges))

    Q1 = np.array([1,0,0])
    Q2 = np.array([0.5,1,0])

    ax,Data,bins = ds.plotCut1D(Q1,Q2,width,rlu=True,minPixel=0.01,Emin=2.0,Emax=2.5,fmt='.')
    Data2,bins2 = ds.cut1D(Q1,Q2,width,rlu=True,minPixel=0.01,Emin=2.0,Emax=2.5)

    assert(Data2.equals(Data.loc[:,Data.columns!='BinDistance']))
    assert(np.all(np.array([np.all(np.isclose(bins[i],bins2[i])) for i in range(len(bins))]).flatten()))

    q1,q2 = ds.convertToQxQy([Q1,Q2])
    D1,b1 = ds.cut1D(Q1,Q2,width,rlu=True,minPixel=0.01,Emin=2.0,Emax=2.5)
    D2,b2 = ds.cut1D(q1,q2,width,rlu=False,minPixel=0.01,Emin=2.0,Emax=2.5)

    # Convert b1 to HKL in order to check if conversion works
    BinPos,OrthoPos,E = b1
    BinPos = np.concatenate([ds.convertToQxQy(BinPos[:,:3]),BinPos[:,-1].reshape(-1,1)],axis=1)
    OrthoPos = ds.convertToQxQy(OrthoPos)
    b1 = [BinPos,OrthoPos,E]

    
    assert(np.all(np.isclose(D1,D2)))
    
    assert(np.all(np.array([np.all(np.isclose(b1[i],b2[i])) for i in range(len(b1))]).flatten()))

    # Check that generating a plot from previous data is equivalent to directly plotting

    
    ax1,cut,bins = ds.plotCut1D(Q1,Q2,width=0.1,minPixel=0.04,Emin=2.0,Emax=2.5,ufit=False)

    ax2,*_ = ds.plotCut1D(Q1,Q2,Emin=2.0,Emax=2.5,width=0.1,minPixel=0.04,data=[cut,bins])



    for key in ['legend','title','xlabel','ylabel']:
        val1 = ax1.properties()[key]
        val2 = ax2.properties()[key]
        assert(val2==val1)

    # check plotted data in line plot 
    # ax1, find index of line2D
    id1 = np.arange(len(ax1.properties()['children']))[np.array([isinstance(child,mpl.lines.Line2D) for child in ax1.properties()['children']])][0]
    id2 = np.arange(len(ax2.properties()['children']))[np.array([isinstance(child,mpl.lines.Line2D) for child in ax2.properties()['children']])][0]

    np.testing.assert_array_equal(ax2.properties()['children'][id1]._xy,ax1.properties()['children'][id2]._xy)


def test_DataSet_1Dcut_ufit():
    q1 =  np.array([1.23,-1.51])
    q2 =  np.array([1.54, -1.25])
    width = 0.1

    plt.ioff()
    import matplotlib
    matplotlib.use('Agg')

    convertFiles = [os.path.join(dataPath,'camea2018n000136.hdf'),os.path.join(dataPath,'camea2018n000137.hdf')]
    
    ds = DataSet(dataFiles = convertFiles)
    ds.convertDataFile(saveFile=False)

    ax,dataset = ds.plotCut1D(q1,q2,width,rlu=False,minPixel=0.01,Emin=2.0,Emax=2.5,fmt='.',ufit=True)
    dataset2 = ds.cut1D(q1,q2,width,rlu=False,minPixel=0.01,Emin=2.0,Emax=2.5,ufit=True)
    

    files = ', '.join([x.replace('hdf','nxs').split(os.path.sep)[-1] for x in convertFiles])

    assert(np.all([np.all(x==y) for x,y in zip(dataset.fit_columns,dataset2.fit_columns)]))
    assert(dataset.meta == dataset2.meta)

    assert(dataset.meta['instrument'] == 'CAMEA')
    assert(dataset.meta['datafilename'] == files)

    

def test_DataSet_1DcutE():
    q =  np.array([1.23,-1.25]).reshape(2,1)
    width = 0.1
    Emin = 1.5
    Emax = 2.5
    plt.ioff()
    import matplotlib
    matplotlib.use('Agg')

    convertFiles = [os.path.join(dataPath,'camea2018n000137.hdf')]
    Datset = DataSet(dataFiles = convertFiles)
    Datset.convertDataFile(saveFile=True)
    Datset._getData()
    I,qx,qy,energy,Norm,Monitor = Datset.I.extractData(),Datset.qx.extractData(),Datset.qy.extractData(),Datset.energy.extractData(),Datset.Norm.extractData(),Datset.Monitor.extractData()

    [intensity,MonitorCount,Normalization,normcounts],[bins] = cut1DE(positions=[qx,qy,energy],I=I,Norm=Norm,Monitor=Monitor,E1=Emin,E2=Emax,q=q,width=width,minPixel=0.01)
    Q = Datset.convertToHKL(q.reshape(2))
    
    Data,[bins] = Datset.cut1DE(E1=Emin,E2=Emax,q=Q,width=width,minPixel=0.01)
    assert(np.min(bins)>=Emin-0.01) # Check that bins do not include data outside of cut
    assert(np.max(bins)<=Emax+0.01)
    assert(len(bins[0])==len(intensity)+1)# Bins denotes edges and must then be 1 more than intensity

    assert(intensity.shape==MonitorCount.shape) # Check that all matrices are cut equally
    assert(intensity.shape==Normalization.shape)
    assert(intensity.shape==normcounts.shape)

    Data,[bins] = Datset.cut1DE(E1=Emin,E2=Emax,q=q,width=width,minPixel=0.01,rlu=False)
    
    Data,[bins] = Datset.cut1DE(E1=Emin,E2=Emax,q=q,width=0.1,minPixel=0.01,rlu=False,constantBins=True)
    
    assert(np.all(np.isclose(np.diff(bins),0.01)))
    assert(bins[0].min()>=Emin)
    assert(bins[0].max()<=Emax)

    try: # no points inside energy interval
        cut1DE(positions=[qx,qy,energy],I=I,Norm=Norm,Monitor=Monitor,E1=500,E2=700,q=q,width=width,minPixel=0.01)
        assert False
    except AttributeError:
        assert True

    try: # no points inside q
        cut1DE(positions=[qx,qy,energy],I=I,Norm=Norm,Monitor=Monitor,E1=5,E2=7,q=np.array([20.0,0]).reshape(2,1),width=width,minPixel=0.01)
        assert False
    except AttributeError:
        assert True

    # Check the data of plot to be the same as cut
    Data,[bins] = Datset.cut1DE(E1=Emin,E2=Emax,q=q,width=0.1,minPixel=0.01,rlu=False,constantBins=True)
    ax,Data2,[bins2] = Datset.plotCut1DE(E1=Emin,E2=Emax,q=q,width=0.1,minPixel=0.01,rlu=False,constantBins=True)
    assert(Data.equals(Data2))
    assert(np.all(np.isclose(bins[0],bins2[0])))

    ufitData = Datset.cut1DE(E1=Emin,E2=Emax,q=q,width=0.1,minPixel=0.01,rlu=False,constantBins=True,ufit=True)
    ax,ufitData2 = Datset.plotCut1DE(E1=Emin,E2=Emax,q=q,width=0.1,minPixel=0.01,rlu=False,constantBins=True,ufit=True)

    files = ', '.join([x.replace('hdf','nxs').split(os.path.sep)[-1] for x in convertFiles])
    
    assert(np.all([np.all(np.isclose(x,y,equal_nan=True)) for x,y in zip(ufitData.fit_columns,ufitData2.fit_columns)]))
    assert(ufitData.meta == ufitData2.meta)

    assert(ufitData.meta['instrument'] == 'CAMEA')
    assert(ufitData.meta['datafilename'] == files)

    ax,Data3,[bins3] = Datset.plotCut1DE(E1=Emin,E2=Emax,q=Q,width=0.1,minPixel=0.01,constantBins=True)



def test_DataSet_2Dcut():
    q1 =  np.array([1.23,-1.25])
    q2 =  np.array([1.54, -1.51])
    width = 0.1
    minPixel=0.02
    EnergyBins = np.linspace(2,3,4)
    plt.ioff()
    import matplotlib
    matplotlib.use('Agg')

    convertFiles = [os.path.join(dataPath,'camea2018n000137.hdf')]

    Datset = DataSet(dataFiles = convertFiles)

    Datset.convertDataFile(saveFile=False)
    ax,Data,bins = Datset.plotCutQE(q1,q2,width=width,minPixel=minPixel,EnergyBins=EnergyBins,rlu=False)# Remove to improve test coverage ,vmin=0.0 , vmax= 5e-06)
    Data2,bins2,_ = Datset.cutQE(q1,q2,width=width,minPixel=minPixel,EnergyBins=EnergyBins,rlu=False)

    comparisons = list(Data.columns)
    del comparisons[comparisons.index('BinDistance')]

    np.all([np.allclose(Data[p],Data2[p],equal_nan=True) for p in comparisons])

    Q1 = np.array([1,0,0])
    Q2 = np.array([0.5,1,0])

    q1,q2 = Datset.convertToQxQy([Q1,Q2])

    Data1,bins,_ = Datset.cutQE(Q1,Q2,width,minPixel,EnergyBins=EnergyBins,rlu=True)
    Data2,bins2,_ = Datset.cutQE(q1,q2,width,minPixel,EnergyBins=EnergyBins,rlu=False)

    comparisons = list(Data.columns)
    del comparisons[comparisons.index('BinDistance')]
    
    assert(np.all(comparisons == Data2.columns))

    
    np.all([np.allclose(Data1[p],Data2[p],equal_nan=True) for p in comparisons])

def test_DataSet_cutPowder():

    plt.ioff()
    import matplotlib
    matplotlib.use('Agg')

    convertFiles = [os.path.join(dataPath,'camea2018n000136.hdf')]
    
    Datset = DataSet(dataFiles = convertFiles)
    Datset.convertDataFile(saveFile=True)
    mask = Datset.mask#np.ones_like(Datset.mask)[0]

    Datset.mask = mask
    Datset.mask = np.logical_not(mask)
    

    eBins = _tools.binEdges(Datset.energy,0.25)
    qBins = np.linspace(0.0,4,201)

    ax,D,QCentres,ECentres = Datset.plotCutPowder(eBins,qBins)# Remove to improve test ,vmin=0,vmax=1e-6)
    D2 = Datset.cutPowder(eBins,qBins)
    assert(np.all(D.equals(D2)))
    #assert(np.all(np.isclose(QCentres2,QCentres)))
    #assert(np.all(np.isclose(ECentres2,ECentres)))

    
def test_DataSet_createRLUAxes():
    plt.ioff()
    import matplotlib
    matplotlib.use('Agg')

    fig = plt.figure()
    convertFiles = [os.path.join(dataPath,'camea2018n000136.hdf')]
    
    ds = DataSet(dataFiles = convertFiles)
    ds.convertDataFile(saveFile=True)

    ax = ds.createQAxis()
    ax = ds.createQAxis(basex=0.5,figure=fig)
    ax = ds.createQAxis(basey=0.5)

    if pythonVersion == 3: # Only possible in python 3
        ax.set_xticks_base(0.2)
        ax.set_yticks_base(0.5)

    V1,V2,V3 = [2,0,0],[-2,3,0],[2,-3,0]
    ax.set_axis(V1,V2)
    ax.set_axis(V1,V2,V3)

    plt.close('all')


def test_DataSet_createQEAxes():
    plt.ioff()
    import matplotlib
    matplotlib.use('Agg')

    convertFiles = [os.path.join(dataPath,'camea2018n000136.hdf')]
    
    ds = DataSet(dataFiles = convertFiles)
    ds.convertDataFile(saveFile=True)

    ax = ds.createQEAxes(projectionVector1=ds.sample[0].projectionVector1,projectionVector2=ds.sample[0].projectionVector2)

    try:
        ax = ds.createQEAxes(axis=2) # Axis only allowed to be 0 or 1
    except AttributeError:
        assert True
    
    try:
        ax = ds.createQEAxes(projectionVector1=[1,0,0],projectionVector2=[1,2,3,4,5]) # Wrong shape of vector
    except AttributeError:
        assert True
    plt.close('all')



def test_DataSet_plotQPlane():
    plt.ioff()
    import matplotlib
    matplotlib.use('Agg')

    convertFiles = [os.path.join(dataPath,'camea2018n000137.hdf')]#'TestData/ManuallyChangedData/A3.hdf')]
    
    Datset = DataSet(dataFiles = convertFiles)
    Datset.convertDataFile(saveFile=True)

    EmptyDS = DataSet()
    try:
        Datset.plotQPlane() # No Bins, Emin or Emax
        assert False
    except AttributeError:
        assert True
    try:
        Datset.plotQPlane(EBins=[10]) # Length of bins is 1
        assert False
    except AttributeError:
        assert True
    
    try:
        Datset.plotQPlane(EMin=20,EMax=10) # EMin>EMax
        assert False
    except AttributeError:
        assert True
    
    try:
        EmptyDS.plotQPlane(EMin=2,EMax=3) # Empty DataSet
        assert False
    except AttributeError:
        assert True


    EMin = Datset.energy.min()
    EMax = EMin+0.5
    Data,ax1 = Datset.plotQPlane(EMin,EMax,binning='xy',xBinTolerance=0.05,yBinTolerance=0.05,enlargen=True,log=False,rlu=True)
    Data,ax2 = Datset.plotQPlane(EMin,EMax,binning='polar',xBinTolerance=0.05,yBinTolerance=0.05,enlargen=False,log=True,rlu=True)
    AX = Datset.createQAxis()
    Data,ax3 = Datset.plotQPlane(EMin,EMax,binning='xy',xBinTolerance=0.05,yBinTolerance=0.05,enlargen=False,ax=AX,colorbar=True,vmin=0,vmax=1e-6,zorder=10)
    
    ax1.set_clim(-20,-15)
    ax2.set_clim(0,1e-6)
    Data,ax3 = Datset.plotQPlane(EMin,EMax,binning='xy',xBinTolerance=0.05,yBinTolerance=0.05)
    
    cmap = plt.cm.coolwarm

    Dataset = DataSet(dataFiles=convertFiles)
    for d in Dataset.dataFiles:
        d.A3Off +=90 # rotate data to fall into problem of arctan2
    Data,ax2 = Datset.plotQPlane(EMin,EMax,binning='polar',xBinTolerance=0.05,yBinTolerance=0.05,enlargen=False,log=True,rlu=True,cmap=cmap)
    
    try:
        Datset.plotQPlane(EMin,EMax,binning='notABinningMethod')
        assert False
    except:
        assert True

    # 3D
    from mpl_toolkits.mplot3d import Axes3D
    from matplotlib.colors import ListedColormap
    cmap = plt.cm.coolwarm
    my_cmap = cmap(np.arange(cmap.N))
    my_cmap[:,-1] = np.linspace(0, 1, cmap.N)
    my_cmap = ListedColormap(my_cmap)

    fig = plt.figure(figsize=(10,11))
    ax = fig.add_subplot(111, projection='3d')

    Energies = np.concatenate(Datset.energy,axis=0)
    E = np.arange(Energies.min()+0.35,Energies.max(),0.35)


    PandasData,ax = \
    Datset.plotQPlane(EBins=E,ax = ax,xBinTolerance=0.03,yBinTolerance=0.03,
            binning='polar',vmin=7.5e-7,vmax=7e-6,antialiased=True,cmap=cmap,rlu=True,extend='max')
    plt.close('all')

#@pytest.mark.unit
#def test_DataSet_plotA3A4(quick):
#    plt.ioff()
#
#    File1 = os.path.join(dataPath,'camea2018n000136.hdf')
#    File2 = os.path.join(dataPath,'camea2018n000137.hdf')
#
#    DS = DataSet(dataFiles=[File1,File2])
#    DS.convertDataFile(saveFile=True)
#
#    F1 = DS.convertedFiles[0]
#    F2 = DS.convertedFiles[1]
#
#    files = [F1,F2]
#    axes = [plt.figure().gca(),plt.figure().gca()]
#    try:
#        plotA3A4(files,planes=[],ax=axes) # 64 planes and only 2 axes
#        assert False
#    except AttributeError:
#        assert True
#
#    try:
#        plotA3A4(files,planes=None,ax=[]) # 64 planes and only 2 axes
#        assert False
#    except AttributeError:
#        assert True 
#
#    try:
#        plotA3A4(files,planes=[[0,2,3],23,44],ax=axes) # 3 planes and 2 axes
#        assert False
#    except AttributeError:
#        assert True
#    
#    try:
#        ei = F1.Ei
#        F2.Ei = F1.Ei*10
#        plotA3A4(files,planes=[[0,2,3],23,44],ax=axes) # 3 planes and 2 axes
#        assert False
#    except AttributeError:
#        F2.Ei = ei
#        assert True
#
#    try:
#        plotA3A4(files,planes=[10,[22]],ax=axes,singleFigure=True) # 2 axes and singleFigure true
#        assert False
#    except AttributeError:
#        assert True
#    if not quick==True:
#        print('___________')
#        plotA3A4(files,planes=[10,[22,23]],ax=axes) # Plot plane 10 and 22+23 in the provided axes
#        print('___________')
#        DS.plotA3A4(planes=[19,[22,25]]) # Plot planes in new axes
#        print('___________')
#        DS.plotA3A4([F1,F1],planes=[19,[22,25]]) # Plot planes in new axes
#        print('___________')
#        patches,energies=DS.plotA3A4([F1],planes=[10,25],returnPatches=True)
#        print('___________')
#        assert(len(patches)==2)
#        assert(len(energies)==2)
#    plt.close('all')

@pytest.mark.unit
def test_DataSet_plotQPatches(quick):
    assert True
    #     plt.ioff()

#     File1 = 'TestData/T0Phonon10meV.nxs')
#     File2 = 'TestData/T0Phonon10meV93_5A4.nxs')

#     DS = DataSet(convertedFiles=[File1,File2])

#     F1 = DS.convertedFiles[0]
#     F2 = DS.convertedFiles[1]

#     files = [F1,F2]
#     axes = [plt.figure().gca(),plt.figure().gca()]
#     try:
#         plotQPatches(files,planes=[],ax=axes) # 64 planes and only 2 axes
#         assert False
#     except AttributeError:
#         assert True
        
#     try:
#         plotQPatches(files,planes=[[0,2,3],23,44],ax=axes) # 3 planes and 2 axes
#         assert False
#     except AttributeError:
#         assert True

#     try:
#         plotQPatches(files,planes=[10,[22]],ax=axes,singleFigure=True) # 2 axes and singleFigure true
#         assert False
#     except AttributeError:
#         assert True

#     if not quick==True:
#         plotQPatches(files,planes=[10,[22,23]],ax=axes) # Plot plane 10 and 22+23 in the provided axes
#         DS.plotQPatches(planes=[19,[22,25]],A4Extend=0.5,A3Extend=1) # Plot planes in new axes
#         DS.plotQPatches(dataFiles=[files[0],files[0]],planes=[19,[22,25]],A4Extend=0.5,A3Extend=1) # Plot planes in new axes and only one file
#     plt.close('all')
    

def test_DataSet_fmt():
    assert('$1.00 \\times 10^{1}$' == fmt(10,'Unused'))
    assert('$1.00 \\times 10^{-10}$' == fmt(1e-10,'Unused'))
    assert('$2.55 \\times 10^{-2}$' == fmt(0.0255,'Unused'))
    assert('$2.56 \\times 10^{-2}$' == fmt(0.02556,'Unused'))
    

def test_DataSet_figureRowColumns():
    assert(np.all(np.array([3,4])==np.array(figureRowColumns(10)))) # 10 -> 3,4
    assert(np.all(np.array([3,3])==np.array(figureRowColumns(9)))) # 9 -> 3,3
    assert(np.all(np.array([1,1])==np.array(figureRowColumns(1)))) # 1 -> 1,1
    try:
        figureRowColumns(0) # No figures
        assert False
    except AttributeError:
        assert True

    assert(np.all(np.array([8,8])==np.array(figureRowColumns(63)))) # 63 -> 8,8
    

def test_DataSet_centeroidnp():
    pos = np.array([[0,0],[1,0],[0,1],[1,1]],dtype=float)
    assert(np.all(np.isclose(np.array([0.5,0.5]),centeroidnp(pos))))

    pos2 = np.array([[1.2,2.2],[7.5,1.0],[11.0,0.0],[4.0,-1.0],[2.0,2.0]],dtype=float)
    assert(np.all(np.isclose(np.array([5.14,0.84]),centeroidnp(pos2))))
    
def test_DataSet_compareNones():
    assert(compareNones(np.array([None]),np.array([None]),0.1))
    assert(not compareNones(np.array([None]),np.array([0.5]),0.1))
    assert(not compareNones(np.array([0.5]),np.array([None]),0.1))
    assert(compareNones(np.array([0.4]),np.array([0.5]),0.2))
    assert(not compareNones(np.array([0.4]),np.array([0.5]),0.001))

    assert(not np.all(compareNones(np.array([0.4,10.2,10.0]),np.array([0.5]),0.001)))
    assert(np.all(compareNones(np.array([0.4,10.2,10.0]),np.array([0.4,10.2,10.0]),0.001)))


def test_DataSet_cutQELine():
    QPoints = np.array([[0.3,-1],[0.7,-1.4],[1.6,-0.9],[0.3,-0.9]],dtype=float)
    QPointsHKL=np.array([[1.0,0.0,0.0],
                        [0.5,1.5,0.0],
                        [1.7,-0.1,0.0],
                        [1.0,1.0,0.0]])


    EnergyBins = np.linspace(1.7,2.7,5)
    minPixel = 0.001
    width=0.1
    DataFile = [os.path.join(dataPath,'camea2018n000137.hdf')]

    dataset = DataSet(convertedFiles=DataFile)
    dataset.convertDataFile(saveFile=False)

    try: # No Q-points
        dataset.cutQELine([],EnergyBins,width=width,minPixel=minPixel,rlu=True)
        assert False
    except AttributeError:
        assert True
    try: # Wrong RLU-input
        dataset.cutQELine([],EnergyBins,width=width,minPixel=minPixel,rlu='42') # Wrong RLU-input
        assert False
    except AttributeError:
        assert True


    DataList,BinList=dataset.cutQELine(QPointsHKL,EnergyBins,width=width,minPixel=minPixel,rlu=True)
    DataList2,BinList2=dataset.cutQELine(QPoints,EnergyBins,width=width,minPixel=minPixel,rlu=False)
    print(DataList)
    assert(len(DataList)==(len(QPointsHKL)-1))# Assert that there are 3 cuts 

    assert(len(DataList2)==(len(QPoints)-1))

    
    

def test_DataSet_plotCutQELine():
    
    Points = np.array([[0.7140393034102988,-0.4959224853328328],
                        [1.128363301356428,-1.6520150761601147],
                        [1.9002545852012716,-0.9393552598967219],
                        [1.0432282332853056,-0.12375569239528339]],dtype=float)
    QPoints = np.zeros((Points.shape[0],3))
    QPoints[:,:2]=Points
    EnergyBins = np.linspace(1.7,2.7,11)
    minPixel = 0.001
    width=0.1
    import matplotlib
    matplotlib.use('Agg')

    DataFile = [os.path.join(dataPath,'camea2018n000136.hdf'),os.path.join(dataPath,'camea2018n000137.hdf')]
    dataset = DataSet(convertedFiles=DataFile)
    dataset.convertDataFile(saveFile=False)
    
    try: # No Q-points
        dataset.plotCutQELine([],EnergyBins,width=width,minPixel=minPixel,rlu=False)
        assert False
    except AttributeError:
        assert True

    try: # No wrong dim of QPonts
        dataset.plotCutQELine(QPoints,EnergyBins,width=width,minPixel=minPixel,rlu=False)
        assert False
    except AttributeError:
        assert True

    try: # No wrong dim of QPonts
        dataset.plotCutQELine(QPoints[:,:2],EnergyBins,width=width,minPixel=minPixel,rlu=True)
        assert False
    except AttributeError:
        assert True


    ax,Data = dataset.plotCutQELine(
        QPoints[:,:2],EnergyBins,width=width,minPixel=minPixel,rlu=False,vmin=0.0,vmax=1.5e-6,log=True,seperatorWidth=3)


    HKLPoints = np.array([[1.0,0.0,0.0],
                        [0.5,1.5,0.0],
                        [1.7,-0.1,0.0],
                        [1.0,1.0,0.0]])



    ax,Data = dataset.plotCutQELine(
        HKLPoints,EnergyBins,width=width,minPixel=minPixel,rlu=True,plotSeperator = False,colorbar=True,log=True)

    
    if True:
        # 3D
        from mpl_toolkits.mplot3d import Axes3D
        from matplotlib.colors import ListedColormap
        cmap = plt.cm.coolwarm
        my_cmap = cmap(np.arange(cmap.N))
        my_cmap[:,-1] = np.linspace(0, 1, cmap.N)
        my_cmap = ListedColormap(my_cmap)

        fig = plt.figure(figsize=(10,11))
        ax = fig.add_subplot(111, projection='3d')

        Energies = np.concatenate(dataset.energy,axis=0)
        E = np.arange(Energies.min()+0.35,Energies.max(),0.35)
        
        try:
            ax,Data,Bins = \
            dataset.plotCutQELine(QPoints=HKLPoints,EnergyBins=E,ax = ax,width=0.05,minPixel=0.01,
                    vmin=7.5e-7,vmax=7e-6,cmap=cmap,rlu=True)
            assert False
        except NotImplementedError:
            assert True



def test_DataSet_extractDetectorData():
    DataFile = [os.path.join(dataPath,'camea2018n000136.hdf'),os.path.join(dataPath,'camea2018n000137.hdf')]#['TestData/ManuallyChangedData/A3.nxs'),'TestData/ManuallyChangedData/A3.nxs')]
    dataset = DataSet(DataFile)

    binning = 1
    dataset.convertDataFile(binning=binning,saveFile=True)

    
    try:
        dataset.extractDetectorData(A4=10000.0) # A4 outside of detector
        assert False
    except AttributeError:
        assert True

    try:
        dataset.extractDetectorData(Ef=10000.0) # Ef outside of detector
        assert False
    except AttributeError:
        assert True
    

    Efs = dataset.convertedFiles[0].instrumentCalibrationEf[:,1].reshape(104,8*binning)
    AnalyserSelection = 5
    Ef = np.mean(Efs[:,AnalyserSelection])

    A4s = dataset.convertedFiles[0].instrumentCalibrationA4.reshape(104,8*binning)
    DetectorSelection = 19
    A4 = np.mean(A4s[DetectorSelection])-dataset.convertedFiles[0].A4Off


    DatBoth = dataset.extractData(A4=A4,Ef=Ef)
    DatBothId = dataset.extractData(A4=A4,EfId=AnalyserSelection)
    DatOne = dataset.extractData(A4=A4)
    DatOne2= dataset.extractData(Ef=Ef)
    DatAll = dataset.extractData()
    DatAllRaw = dataset.extractData(raw=True)


    # Independent of number of files:
    assert(len(DatAllRaw)==3) # Check that 3 lists are returned
    assert(len(DatAllRaw[0])==len(DatAllRaw[1]) and len(DatAllRaw[0])==len(DatAllRaw[2])) # Check that 3 list have same number of files

    assert(np.all(DatBothId[0]==DatBoth[0])) # Check that ID and value gives the same.

    # The shape of raw is the same as non-raw
    assert(len(DatAllRaw[0])==len(DatAll)) # Have same number of files

    for i in range(len(dataset.convertedFiles)):
        assert(DatAllRaw[0][i].shape==DatAllRaw[1][i].shape and DatAllRaw[0][i].shape==DatAllRaw[2][i].shape) # Check that 3 list have same shape
        assert(DatAllRaw[0][i].shape==DatAll[i].shape) 
        

def test_DataSet_subract():
    #Simple test of subtracting the same data file from it-self
    DataFile = [os.path.join(dataPath,'camea2018n000136.hdf'),os.path.join(dataPath,'camea2018n000137.hdf')]
    dataset = DataSet(DataFile)

    dataset2 = DataSet(DataFile)

    try:
        subtracted = dataset-dataset2
    except AttributeError: # Data set need to be converted
        assert True
    
    dataset.convertDataFile()
    dataset2.convertDataFile()
    subtracted = dataset-dataset2
    assert(np.sum(subtracted.I.data)==0)
    assert(np.all(np.isclose(subtracted.Norm.data,dataset.Norm.data)))
    assert(np.all(np.isclose(subtracted.Monitor.data,dataset.Monitor.data)))

def test_DataSet_OxfordList():
    l = ['Apples','Pears']
    S = OxfordList(l)
    assert(S=='Apples and Pears')

    l.append('Oranges')
    S = OxfordList(l)
    assert(S=='Apples, Pears, and Oranges')

    assert(OxfordList([]) is None)
    assert(OxfordList(['Apples'])=='Apples')

@pytest.mark.skip(reason="Current data structure not defined")
def test_DataSet_MultiFLEXX():
    fileLocation = _tools.fileListGenerator('65059',folder=os.path.join('Data',''),instrument='MultiFLEXX')

    ds = DataSet(fileLocation)
    ds.convertDataFile(saveFile = False)
    import matplotlib
    matplotlib.use('Agg')

    V = ds.View3D(0.05,0.05,0.5,grid=True)

def test_DataSet_ELine():
    DataFile = [os.path.join(dataPath,'camea2018n000136.hdf'),os.path.join(dataPath,'camea2018n000137.hdf')]
    dataset = DataSet(DataFile)

    dataset.convertDataFile()
    import matplotlib
    matplotlib.use('Agg')

    Q1 = [1.0,-0.185,0.0]
    Q2 = [0.5,1.5,0.0]
    Emin = 1.65
    Emax = 3.3

    CutData,Bins = dataset.cutELine(Q1, Q2, Emin=Emin, Emax=Emax, energyWidth = 0.05, minPixel = 0.02, width = 0.02, rlu=True, dataFiles=None, constantBins=False)
    ax, CutDataPlot, BinsPlot = dataset.plotCutELine(Q1, Q2, Emin=Emin, Emax=Emax, energyWidth = 0.05, minPixel = 0.02, width = 0.02, rlu=True, dataFiles=None, constantBins=False)

    assert(np.all([np.all(np.isclose(B,B2)) for B,B2 in zip(Bins,BinsPlot)]))
    assert(np.all(CutDataPlot.equals(CutData)))

    assert(np.all(np.logical_and(CutData['Energy']>=Emin,CutData['Energy']<=Emax*1.01))) # Allow for slightly heigher energy
    assert(np.logical_and(np.all(CutData['H']<=Q1[0]*1.01),np.all(CutData['H']>=Q2[0]*0.99)))
    assert(np.logical_and(np.all(CutData['K']>=Q1[1]*1.01),np.all(CutData['K']<=Q2[1]*1.01)))
    assert(np.all(np.isclose(CutData['L'],0.0,atol=1e-6)))
    assert(np.all([np.all(np.logical_and(B[0]>=Emin*0.99,B[0]<=Emax*1.05)) for B in Bins])) # Allow for slightly heigher energy


    if sys.version[0] == '3':
        ax.set_xticks_base(0.01)
        ax.set_xticks_number(10)
        ax.set_xticks_base()
        

    Q1raw = dataset.convertToQxQy(Q1)
    Q2raw = dataset.convertToQxQy(Q2)

    CutData,Bins = dataset.cutELine(Q1raw, Q2raw, Emin=Emin, Emax=Emax, energyWidth = 0.05, minPixel = 0.02, width = 0.02, rlu=False)
    ax, CutDataPlot, BinsPlot = dataset.plotCutELine(Q1raw, Q2raw, Emin=Emin, Emax=Emax, energyWidth = 0.05, minPixel = 0.02, width = 0.02, rlu=False)


    assert(np.all([np.all(np.isclose(B,B2)) for B,B2 in zip(Bins,BinsPlot)]))
    assert(np.all(CutDataPlot.equals(CutData)))

    assert(np.all(np.logical_and(CutData['Energy']>=Emin,CutData['Energy']<=Emax*1.01))) # Allow for slightly heigher energy
    assert(np.logical_and(np.all(CutData['Qx']<=Q1raw[0]*1.01),np.all(CutData['Qx']>=Q2raw[0]*0.99)))
    assert(np.logical_and(np.all(CutData['Qy']<=Q1raw[1]*0.99),np.all(CutData['Qy']>=Q2raw[1]*1.01)))

    assert(np.all([np.all(np.logical_and(B[0]>=Emin*0.99,B[0]<=Emax*1.05)) for B in Bins])) # Allow for slightly heigher energy


def test_updateCalibration():
    calibFiles = [os.path.join('Data','Normalization80_1.calib'),
                    os.path.join('Data','Normalization80_3.calib'),
                    os.path.join('Data','Normalization80_5.calib')]


    ds = DataSet(os.path.join(dataPath,'camea2018n000136.hdf'))
    
    df = ds[0]

    df.loadBinning(1)

    binnings = df.possibleBinnings # is 1,3,8
    edges = df.instrumentCalibrationEdges

    ds.updateCalibration(calibFiles)

    df.loadBinning(1)
    newBinnings = df.possibleBinnings # is 1,3,8,5
    newEdges = df.instrumentCalibrationEdges
    assert(len(newBinnings)!=len(binnings)) # Addition of binning 5
    assert(not np.any(newEdges!=edges)) # Check if all elemenst are equal


    ds.updateCalibration(calibFiles,overwrite=True)
    df.loadBinning(1)

    newEdges = df.instrumentCalibrationEdges
    assert(np.any(newEdges!=edges)) # Check if all elemenst are equal


def testplotRaw1D_Error():
    DataFile = [os.path.join(dataPath,'camea2018n000136.hdf'),os.path.join(dataPath,'camea2018n000137.hdf')]
     # Scan variables are A3 and A3+A4
    dataset = DataSet(DataFile)
    dataset[0].scanParameters = ['Ei']
    import matplotlib
    matplotlib.use('Agg')

    try:
        dataset.plotRaw1D() # Two different scan types
        assert False
    except AttributeError:
        assert True
        
    DataFile = [os.path.join(dataPath,'camea2018n000137.hdf'),os.path.join(dataPath,'camea2018n000137.hdf')]
    dataset = DataSet(DataFile)
    try:
        dataset.plotRaw1D(analyzerSelection=[0]) # Not enough analyzers provided
        assert False
    except AttributeError:
        assert True
    try:
        dataset.plotRaw1D(detectorSelection=[0]) # Not enough detectors provided
        assert False
    except AttributeError:
        assert True
        
    try:
        dataset.plotRaw1D(legend=[0]) # Not enough legend labels provided
        assert False
    except AttributeError:
        assert True
        
def testplotRaw1D():
    DataFile = [os.path.join(dataPath,'camea2018n000137.hdf'),os.path.join(dataPath,'camea2018n000137.hdf')]
    dataset = DataSet(DataFile)
    import matplotlib
    matplotlib.use('Agg')

    ax = dataset.plotRaw1D()
    
    ax = dataset.plotRaw1D(legend=['1','2'],detectorSelection=[0,0],analyzerSelection=[5,5],grid=True)
    assert True

def testMasking():
    DataFile = [os.path.join(dataPath,'camea2018n000136.hdf'),os.path.join(dataPath,'camea2018n000137.hdf')]
     # Scan variables are A3 and A3+A4
    ds = DataSet(DataFile)
    ds.convertDataFile()

    circ = Mask.circleMask(center=np.array([1.0,0.0]),radiusPoint=np.array([1.1,0.0]),coordinates =['h','k'])
    Emask = Mask.lineMask(1.7,end=2.0,coordinates='energy')
    rect = Mask.rectangleMask(corner1=np.array([1.0,0.0]),corner2=np.array([1.5,0.5]),coordinates=['h','k'])

    mask = circ*Emask+Emask*rect
    calcMask = (circ*Emask)(ds)
    ds.mask = mask(ds)
    masks = [mask(df) for df in ds]
    
    assert(np.sum(ds.I.mask)==np.sum([np.sum(m) for m in masks]))


    # Test failing mask
    try:
        ds.mask = 'Error'
        assert False
    except AttributeError:
        assert True
    
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        mask = np.array([np.ones_like(I) for I in ds.I]) # Mask all
        ds.mask = mask

    ds = DataSet([DataFile[0],DataFile[0]])
    ds.convertDataFile()

    
    print([x.shape for x in mask])
    print(mask[0].shape)
    
    ds.mask = mask[0] # Only provide 1 mask to be applied to both data files
    
    

def testupdateSampleParameters():
    DataFile = [os.path.join(dataPath,'camea2018n000136.hdf'),os.path.join(dataPath,'camea2018n000137.hdf')]
    ds = DataSet(DataFile)

    unitCell = ds.sample[0].unitCell
    unitCell[:2]*=2.0
    ds.updateSampleParameters(unitCell=unitCell)
    newCell = ds.sample[0].cell
    newUB = ds.sample[0].UB

    for d in ds:
        assert(np.all(np.isclose(d.sample.cell,newCell)))
        assert(np.all(np.isclose(d.sample.UB,newUB)))

def test_CurratAxeMasking():
    DataFile = [os.path.join(dataPath,'camea2018n000136.hdf'),os.path.join(dataPath,'camea2018n000137.hdf')]
    ds = DataSet(DataFile)
    
    # Correct wrong alignment of about 3.5 degrees
    theta=np.rad2deg(0.05979407243252655)
    for df in ds:
        df.A3-=theta
    ds.convertDataFile(binning=3)

    averageSum = np.mean(ds.I.extractData())

    braggpeaks = [[1,0,0]]
    # Calcualte mask directly from dataset
    curratAxeMask = ds.calculateCurratAxeMask(braggpeaks,dH=0.1,dK=0.1,dL=0.001)
    ds.mask = curratAxeMask

    # Calcualte mask using a mask object and on the dataset
    CAMask = Mask.CurratAxeMask(braggpeaks,dH=0.1,dK=0.1,dL=0.001)

    evaluatedCAMask = CAMask(ds)
    assert(np.all(np.equal(evaluatedCAMask,curratAxeMask)))

    # Calcualte the two different contributions
    CAMaskMono = Mask.CurratAxeMask(braggpeaks,dH=0.1,dK=0.1,dL=0.001,spurionType='Monochromator')
    CAMaskAna = Mask.CurratAxeMask(braggpeaks,dH=0.1,dK=0.1,dL=0.001,spurionType='Analyser')
    CAMaskMonoEvaluated = CAMaskMono(ds)
    CAMaskAnaEvaluated = CAMaskAna(ds)
    CAMaskCombi = [np.logical_or(CAMM,CAMA) for CAMM,CAMA in zip(CAMaskMonoEvaluated,CAMaskAnaEvaluated)]
    assert(np.all(np.equal(evaluatedCAMask,CAMaskCombi)))


    averageSumMasked = np.mean(ds.I.extractData())
    assert(averageSum-averageSumMasked>0.004)

    curratAxeMask = ds.calculateCurratAxeMask(braggpeaks,dqx=0.15,dqy=0.15)
    ds.mask = curratAxeMask


    averageSumMasked = np.mean(ds.I.extractData())
    assert(averageSum-averageSumMasked>0.004)
    


def test_absoluteNormalziation():
    DataFile = [os.path.join(dataPath,'camea2018n000136.hdf')]
    ds = DataSet(DataFile)

    try:
        ds.absoluteNormalize(10.0,'MnF2')
        assert False
    except  AttributeError: # Must be converted first!
        assert True

    ds.convertDataFile()
    norm = np.mean(ds.Norm.extractData())
    
    # Use value for MnF2 to check 
    ds.absoluteNormalize(sampleMass=6.2,sampleChemicalFormula='MnF2',formulaUnitsPerUnitCell=2,
                                      correctVanadium=False)
    
    factor = 0.06088201383247563 # Factor calculated for MnF2

    assert(np.isclose(factor,ds.absoluteNormalized))

    # Redo normalization to retrive the same factor
    ds.absoluteNormalize(sampleMass=6.2,sampleChemicalFormula='MnF2',formulaUnitsPerUnitCell=2,
                                      correctVanadium=False)

    assert(np.isclose(factor,ds.absoluteNormalized))

def test_CustomAxisInput():
    # dset object generation
    DataFile = [os.path.join(dataPath,'camea2018n000136.hdf'),
                os.path.join(dataPath,'camea2018n000136.hdf')]
    ds = DataSet(DataFile)
    ds.convertDataFile()
    # parameter definition for plots
    cutqe_params = {
        "q1":np.array([0.75,0,0], float),
        "q2":np.array([1,0,0], float),
        "EMin":1.8, "EMax":3.6, "dE":0.1, "width":0.1, "minPixel":0.05,
        "rlu":True, "smoothing":0.0, "vmin":0.0, "vmax":1e-07
    }

    # custom axes generation
    fig, axes = plt.subplots(2,1) # two axes stacked on top of each other in same fig
    ax_direction1, ax_direction2 = axes.flatten()

    # plotting and cutting with MJOLNIR - fixed now for QE cuts
    q2 = [1,1,0]
    cutqe_params.update({"q2":q2, "ax":ax_direction1})
    ax_direction1, data_dir1, bins_dir1 = ds.plotCutQE(**cutqe_params)

    q2 = np.array([0.7,0.7,0],float)
    cutqe_params.update({"q2":q2, "ax":ax_direction2})
    ax_direction2, data_dir2, bins_dir2 = ds.plotCutQE(**cutqe_params)

    ## Same for plotCutQELine
    cutqe_params = {
        "QPoints":np.array([[0.75,0,0],[1,0,0]], float),
        'EnergyBins':np.arange(1.8,3.6,0.1), "width":0.1, "minPixel":0.05,
        "rlu":True, "vmin":0.0, "vmax":1e-07
    }

    # custom axes generation
    fig, axes = plt.subplots(2,1) # two axes stacked on top of each other in same fig
    ax_direction1, ax_direction2 = axes.flatten()

    # plotting and cutting with MJOLNIR - fixed now for QE cuts

    cutqe_params.update({"ax":ax_direction1})
    ax_direction1, data_dir1 = ds.plotCutQELine(**cutqe_params)

    QPoints = np.array([[1,0,0],[1,1,0]], float)
    cutqe_params.update({'QPoints':QPoints, "ax":ax_direction2})
    ax_direction2, data_dir2 = ds.plotCutQELine(**cutqe_params)

    ## cut1D
    cutqe_params = {
            "q1":np.array([0,0,0], float),
            "q2":np.array([1,0,0], float),
            "Emin":1.8, "Emax":2.0, "width":0.1, "minPixel":0.05,
            "rlu":True, 
        }

    # custom axes generation
    fig, axes = plt.subplots(2,1) # two axes stacked on top of each other in same fig
    ax_direction1, ax_direction2 = axes.flatten()

    # plotting and cutting with MJOLNIR - fixed now for QE cuts

    cutqe_params.update({"ax":ax_direction1})
    ax_direction1, data_dir1, bins_dir1 = ds.plotCut1D(**cutqe_params)

    q2 = np.array([0.7,0.7,0],float)
    cutqe_params.update({'q2':q2, "ax":ax_direction2})
    ax_direction2, data_dir2, bins_dir2 = ds.plotCut1D(**cutqe_params)


    # cut1DE
    cut1de_params = {
            "q":np.array([1,0,0], float),
            "E1":1.8, "E2":3.6, "width":0.5, "minPixel":0.05,
            "rlu":True, 
        }

    # custom axes generation
    fig, axes = plt.subplots(2,1) # two axes stacked on top of each other in same fig
    ax_direction1, ax_direction2 = axes.flatten()

    # plotting and cutting with MJOLNIR - fixed now for QE cuts

    cut1de_params.update({"ax":ax_direction1})
    ax_direction1, data_dir1, bins_dir1 = ds.plotCut1DE(**cut1de_params)

    q = np.array([0.7,0.7,0],float)
    cut1de_params.update({'q':q, "ax":ax_direction2})
    ax_direction2, data_dir2, bins_dir2 = ds.plotCut1DE(**cut1de_params)


def test_symmetrize():

    DataFile = [os.path.join(dataPath,'camea2018n000136.hdf'),]
    ds = DataSet(DataFile)
    ds.convertDataFile()

    qxOriginal = ds[0].qx
    qyOriginal = ds[0].qy

    def identity(H,K,L):
        return H,K,L

    ds.symmetrize(identity)

    qx = ds[0].qx
    qy = ds[0].qy


    assert(np.all(np.isclose(qx,qxOriginal,atol=1e-6)))
    assert(np.all(np.isclose(qy,qyOriginal,atol=1e-6)))

    def absolute(H,K,L):
        return np.abs(H),np.abs(K),np.abs(L)

    ds.symmetrize(absolute)

    assert(np.all(ds[0].h.min()>=0))
    assert(np.all(ds[0].k.min()>=0))
    assert(np.all(ds[0].l.min()>=0))


def test_sorting():

    DataFile = [os.path.join(dataPath,'camea2018n000136.hdf'),os.path.join(dataPath,'camea2018n000137.hdf')]
    ds = DataSet(DataFile)
    ds.convertDataFile()

    ds2 = DataSet([DataFile[1],DataFile[0]])
    ds2.convertDataFile()

    names = [df.name for df in ds]
    names2 = [df.name for df in ds2]

    assert(names[0] == names2[1] and names[1] == names2[0])

    idx = ds.argAutoSort()
    ds.autoSort()

    for idx1,df in zip(idx,ds):
        oldName = names[idx1]
        newName = df.name
        assert(oldName==newName)

    idxes = ds2.argSortTo(ds)
    ds2.applySorting(idxes)
    names = [df.name for df in ds]
    names2 = [df.name for df in ds2]

    assert(names[0] == names2[0] and names[1] == names2[1])

    



collect_ignore_glob = ["MJOLNIR/CommandLineScripts/MJOLNIRHistory.py","MJOLNIR/CommandLineScripts/MJOLNIR3DView.py","MJOLNIR/CommandLineScripts/MJOLNIRCalibrationInspector.py","MJOLNIR/CommandLineScripts/MJOLNIRConvert.py","MJOLNIR/Statistics/CutObject.py"]


import copy
import functools
import os
import sys
from operator import sub

pythonVersion = sys.version_info[0]

sys.path.append('.')
sys.path.append('..')
sys.path.append('../..')

import matplotlib.pyplot as plt
#if pythonVersion == 3: # Only for python 3
import matplotlib.ticker as mticker

import numpy as np
from MJOLNIR import _tools
from matplotlib.patches import Ellipse
from mpl_toolkits.axisartist import SubplotHost#,axislines
import warnings
try:
    from mpl_toolkits.axisartist.grid_helper_curvelinear import \
        GridHelperCurveLinear
except ImportError:
    from mpl_toolkits.axisartist.floating_axes.grid_helper_curvelinear import \
        GridHelperCurveLinear



def calculateBase(l,span,ticks):
    """Calcualte the tick mark base suitable for current span and ticks

    Args:

        - l (grid locator): Matplotlib grid locator

        - span (float): Width of view

        - ticks (int): Number of ticks wanted

    Returns:

        - base (float): Closest base number according to l.multiples
    """
    ytickorder = np.ceil(np.log10(span/ticks))
    minimalMultiplesy = np.argmin(np.abs(np.power(10,-ytickorder)*span/ticks-l.multiples))
    return l.multiples[minimalMultiplesy]*np.power(10,ytickorder)
    


def updateAxisDecorator(ax,direction='both'):
    def axisDecorator(func):
        @functools.wraps(func)
        def newFunc(*args,**kwargs):
            returnval = func(*args,**kwargs)
            ax.axisChanged(direction=direction)
            return returnval
        return newFunc
    return axisDecorator

def updateXAxisDecorator(ax):
    def axisDecorator(func):
        @functools.wraps(func)
        def newFunc(*args,**kwargs):
            returnval = func(*args,**kwargs)
            ax.xAxisChanged()
            return returnval
        return newFunc
    return axisDecorator

class MaxNLocator(mticker.MaxNLocator):
    def __init__(self, nbins=10, steps=None,
                trim=True,
                integer=False,
                symmetric=False,
                prune=None):
        # trim argument has no effect. It has been left for API compatibility
        super(MaxNLocator,self).__init__(nbins, steps=steps,
                                    integer=integer,
                                    symmetric=symmetric, prune=prune)
        self.create_dummy_axis()
        self._factor = 1#0.0#None

    def __call__(self, v1, v2): # pragma: no cover
        if self._factor is not None:
            self.set_bounds(v1*self._factor, v2*self._factor)
            locs = mticker.MaxNLocator.__call__(self)
            return np.array(locs), len(locs), self._factor
        else:
            self.set_bounds(v1, v2)
            locs = mticker.MaxNLocator.__call__(self)
            return np.array(locs), len(locs)

    def set_factor(self, f):
        self._factor = f


class MultipleLocator(mticker.MultipleLocator):
    def __init__(self,base=None):
        if base is None:
            base = 0.25
        super(MultipleLocator,self).__init__(base)
        self.create_dummy_axis()
        self._factor = 1#0.0
        self._multiplerVals = np.array([1,2,4,5,10])
        self.multiples = 1.0/self.multiplerVals

    @property
    def multiplerVals(self):
        return self._multiplerVals

    @multiplerVals.getter
    def multiplerVals(self):
        return self._multiplerVals

    @multiplerVals.setter
    def multiplerVals(self,multiplerVals):
        self._multiplerVals = multiplerVals
        self.multiples = 1.0/multiplerVals


    def __call__(self, v1, v2): # pragma: no cover
        if self._factor is not None:
            self.axis.set_view_interval(v1*self._factor, v2*self._factor)
            locs = mticker.MultipleLocator.__call__(self)
            return np.array(locs), len(locs), self._factor
        else:
            self.axis.set_view_interval(v1, v2)
            locs = mticker.MultipleLocator.__call__(self)
            return np.array(locs), len(locs), None

    def set_factor(self, f):
        self._factor = f

def forceGridUpdate(self):
    self._grid_helper._force_update = True
    self.pchanged()
    self.stale = True

def get_aspect(ax):
    figW, figH = ax.get_figure().get_size_inches()
    _, _, w, h = ax.get_position().bounds
    disp_ratio = (figH * h) / (figW * w)
    data_ratio = sub(*ax.get_ylim()) / sub(*ax.get_xlim())
    return (disp_ratio / data_ratio)

def axisChanged(event,axis=None,forceUpdate=False,direction='both'):
    """Function to recalculate the base number for RLU axis"""
    # Two arguments are needed; event for canvas callback and axis is needed for ax.axisChanged
    if axis is None:
        axis = event
    s = axis.sample
    xlim = axis.get_xlim()
    ylim = axis.get_ylim()
    xlimDiff = np.diff(xlim)
    ylimDiff = np.diff(ylim)
    
    if direction == 'both':
        direction = ['x','y']
        difference = [xlimDiff,ylimDiff]
        locators = [axis._grid_helper.grid_finder.grid_locator1,axis._grid_helper.grid_finder.grid_locator2]
    else:
        if direction == 'x':
            difference = [xlimDiff]  
            locators = [axis._grid_helper.grid_finder.grid_locator1]
        else:
            difference = [ylimDiff]  
            locators = [axis._grid_helper.grid_finder.grid_locator2]
        direction = [direction]
    for d,diff,locator in zip(direction,difference,locators):
        forceUpdate = True
        if isinstance(locator,MultipleLocator):
            if not np.isclose(getattr(axis,'_old{}limDiff'.format(d.capitalize())),diff) \
            or forceUpdate:# if new image size

                points = np.array(np.meshgrid(xlim,ylim)).reshape(-1,2) # Generate the 4 corner points
                Qs = np.array([s.tr(p[0],p[1]) for p in points])
                if d=='x':
                    span = np.max(Qs[:,0])-np.min(Qs[:,0])
                else:
                    span = np.max(Qs[:,1])-np.min(Qs[:,1])
                if not hasattr(axis,'{}ticks'.format(d)):
                    setattr(axis,'{}ticks'.format(d),7)
                
                ticks = getattr(axis,'{}ticks'.format(d))
                setattr(axis,'_old{}limDiff'.format(d.capitalize()),diff)
            

            if not hasattr(axis,'{}base'.format(d)):
                base = calculateBase(locator,span,ticks)
            else:
                base = getattr(axis,'{}base'.format(d))

            locator.set_params(base=base)
        
        elif isinstance(locator,MaxNLocator):
            if hasattr(axis,'{}ticks'.format(d)):
                ticks = getattr(axis,'{}ticks'.format(d))
            else:
                ticks = 7
            locator.set_params(nbins = ticks)
        else:
            return 
    
    # force an update
    axis.forceGridUpdate()



def createQAxis(self,rlu = True, withoutOnClick = False, figure=None,ids=[1, 1, 1],basex=None,basey=None):#,ax=None):
    """Create a reciprocal lattice plot for a given DataSet object.
    
    Args:
        
        - Dataset (DataSet): DataSet object for which the RLU plot is to be made.

    Kwargs:

        - rlu (bool): If true, plot in reciprocal lattice units, else in Qx, Qy (default True)

        - withoutOnClick (bool): If False, utilize the updated onClick method (default False)

        - figure: Matplotlib figure in which the axis is to be put (default None)

        - ids (array): List of integer numbers provided to the SubplotHost ids attribute (default [1,1,1])

        - basex (float): Ticks are positioned at multiples of this value along x (default None)

        - basey (float): Ticks are positioned at multiples of this value along y (default None)

    Returns:
        
        - ax (Matplotlib axes): Axes containing the RLU plot.

    .. note::
        When rlu axis is created, the orientation of Qx and Qy is assumed to be rotated as well. 
        This is to be done in the self.View3D method call!

    .. note::
        When using python 2 the changing of tick marks is not supported due to limitations in matplotlib. However, if python 3 is used, the number 
        of ticks and their location can be change after the initialization using the set_xticks_number, set_yticks_number chaning the wanted number 
        of tick marks, or the set_xticks_base or set_yticks_base to change the base number, see RLU tutorial under Tools. As default a sufficient base
        number is found and will update when zooming.
        
    """
    
    sample = copy.deepcopy(self.sample)
    if rlu:
        for samp in sample:
            samp.convert = np.einsum('ij,j...->i...',samp.RotMat,samp.convert)
            #sample.convert = np.einsum('ij,j...->i...',sample.RotMat,sample.convert)
            samp.convertinv = np.linalg.inv(samp.convert) # Convert from Qx, Qy to projX, projY

            samp.orientationMatrix = np.dot(samp.RotMat3D,samp.orientationMatrix)
            samp.orientationMatrixINV = np.linalg.inv(samp.orientationMatrix)
            samp.theta = 0.0

    if figure is None:# and ax is None:
        fig = plt.figure(figsize=(7, 4))
    else:
        fig = figure#ax.get_figure()
    if rlu:
        def calculateTicks(ticks,angle,round=True):
            val = ticks/np.tan(angle/2.0)
            if round:
                return np.array(np.round(val),dtype=int)
            else:
                return val

        
        if  not basex is None or not basey is None: # Either basex or basey is provided (or both)
            if basex is None:
                basex = calculateTicks(basey,sample[0].projectionAngle,round=False)
            elif basey is None:
                basey = basex/calculateTicks(1.0,sample[0].projectionAngle,round=False)

            grid_locator1 = MultipleLocator(base=basex)
            grid_locator2 = MultipleLocator(base=basey)
        else:
            basex = 0.5
            basey = 0.5

            grid_locator1 = MultipleLocator(base=basex)
            grid_locator2 = MultipleLocator(base=basey)
            
        grid_helper = GridHelperCurveLinear((sample[0].inv_tr, sample[0].tr),grid_locator1=grid_locator1,grid_locator2=grid_locator2)
        

        #if not ax is None:
        #    if not isinstance(ax,axislines.Axes):
        #        subargs=ax.get_subplotspec()
        #        fig.delaxes(ax)
        #        ax = SubplotHost(fig, subargs, grid_helper=grid_helper)
        #        
        #else:
        #    #ax2=ax1
        ax = SubplotHost(fig, *ids, grid_helper=grid_helper)
        #fig.add_subplot(ax)
    else:
        ax = fig.add_subplot(111)
    ax.sample = sample[0]
    ax.rlu = rlu
    
    
    if ax.rlu:
        
        ax.basex = basex
        ax.basey = basey

        def set_axis(ax,v1,v2,*args):
            if args != ():
                points = np.concatenate([[v1,v2],[x for x in args]],axis=0)
            else:
                points = np.array([v1,v2])
                
            if points.shape[1] == 3:
                points = ax.sample.calculateHKLtoProjection(points[:,0],points[:,1],points[:,2]).T
            boundaries = np.array([ax.sample.inv_tr(x[0],x[1]) for x in points])
            ax.set_xlim(boundaries[:,0].min(),boundaries[:,0].max())
            ax.set_ylim(boundaries[:,1].min(),boundaries[:,1].max())
            if pythonVersion == 3: # Only possible in python 3
                ax.forceGridUpdate()


        fig.add_subplot(ax)
        ax.set_aspect(1.)
    ax.grid(True, zorder=0)
    
    if ax.rlu:
        if not np.isclose(ax.sample.projectionAngle,np.pi/2.0,atol=0.001):
            ax.axis["top"].major_ticklabels.set_visible(True)
            ax.axis["right"].major_ticklabels.set_visible(True)

    if rlu:
        ax.format_coord = ax.sample.format_coord
    else:
        def format_coord(x,y):
            return ', '.join([label+' = {:.3f}'.format(val) for val,label in zip([x,y],['qx','qy'])])

        ax.format_coord = format_coord
    
    
    if ax.rlu:
        ax.set_axis = lambda v1,v2,*args: set_axis(ax,v1,v2,*args)

        def beautifyLabel(vec):
            Vec = [x.astype(int) if np.isclose(x.astype(float)-x.astype(int),0.0) else x.astype(float) for x in vec]
            return '{} [RLU]'.format(', '.join([str(x) for x in Vec]))

        ax.set_xlabel(beautifyLabel(ax.sample.projectionVector1))
        ax.set_ylabel(beautifyLabel(ax.sample.projectionVector2))

    else:
        ax.set_xlabel('Qx [1/AA]')
        ax.set_ylabel('Qy [1/AA]')
    
    def onclick(ax, event,Qx,Qy,data,outputFunction): # pragma: no cover
        if event.xdata is not None and ax.in_axes(event):
            try:
                C = ax.get_figure().canvas.cursor().shape() # Only works for pyQt5 backend
            except:
                pass
            else:
                if C != 0:# or ax.drawState != States.INACTIVE:
                    return
        if not ax.in_axes(event):
            return
        printString = ''
        printString+=ax.format_coord(event.xdata, event.ydata)+', '
        QX = np.array(Qx[0])
        QY = np.array(Qy[0])
        
        Qx = 0.25*(QX[1:,1:]+QX[:-1,:-1]+QX[1:,:-1]+QX[:-1,1:])
        Qy = 0.25*(QY[1:,1:]+QY[:-1,:-1]+QY[1:,:-1]+QY[:-1,1:])

        arg = np.argmin(np.linalg.norm(np.array([Qx,Qy])-np.array([event.xdata,event.ydata]).reshape(2,1,1),axis=0))
        arg2D = np.unravel_index(arg,Qx.shape)
        
        cts = data[0][0][arg2D[0],arg2D[1]]
        Norm = data[1][0][arg2D[0],arg2D[1]]
        Mon = data[2][0][arg2D[0],arg2D[1]]
        NC = data[3][0][arg2D[0],arg2D[1]]
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            Intensity = np.divide(cts*NC,Norm*Mon)

        localSize = np.linalg.norm(np.array([QX[arg2D[0],arg2D[1]]-Qx[arg2D[0],arg2D[1]],QY[arg2D[0],arg2D[1]]-Qy[arg2D[0],arg2D[1]]]))

        if not np.isfinite(Intensity) or np.linalg.norm(np.array([Qx[arg2D[0],arg2D[1]]-event.xdata,Qy[arg2D[0],arg2D[1]]-event.ydata]))>localSize:
            printString+='I = NaN'
        else:
            printString+='I = {:.4E}'.format(Intensity)
            printString+=', Cts = {:d}, Norm = {:.3f}, Mon = {:d}, NormCount = {:d}'.format(cts,Norm,int(Mon),NC)
        if not ax.suppressPrint:
            outputFunction(printString)

    if not withoutOnClick:
        ax.onClick = lambda x: onclick(ax,x,ax.Qx,ax.Qy,[ax.intensity,ax.monitorCount,ax.Normalization,ax.NormCount],outputFunction=ax.outputFunction)
        ax._button_press_event = ax.figure.canvas.mpl_connect('button_press_event', ax.onClick)
    


    if ax.rlu: # 
        ax.calculateTicks = lambda value:calculateTicks(value,ax.sample.projectionAngle)
        ax.forceGridUpdate = lambda:forceGridUpdate(ax)
        ax._oldXlimDiff = np.diff(ax.get_xlim())
        ax._oldYlimDiff = np.diff(ax.get_ylim())

        ax.get_aspect_ratio = lambda: get_aspect(ax)

        ax.callbacks.connect('xlim_changed', axisChanged)
        ax.callbacks.connect('ylim_changed', axisChanged)
        ax.get_figure().canvas.mpl_connect('draw_event',lambda event: axisChanged(event,axis=ax,forceUpdate=True))
        ax.axisChanged = lambda direction='both': axisChanged(None,ax,forceUpdate=True,direction=direction)
    
        @updateAxisDecorator(ax=ax,direction='x')
        def set_xticks_base(xBase,ax=ax):
            """Setter of the base x ticks to be used for plotting

            Args:

                - xBase (float): Base of the tick marks

            """
            if not isinstance(ax._grid_helper.grid_finder.grid_locator1,MultipleLocator):
                l1 = MultipleLocator(base=xBase)
                ax._grid_helper.update_grid_finder(grid_locator1=l1)

            ax.xbase = xBase

        @updateAxisDecorator(ax=ax,direction='y')
        def set_yticks_base(yBase,ax=ax):
            """Setter of the base y ticks to be used for plotting

            Args:

                - yBase (float): Base of the tick marks

            """
            if not isinstance(ax._grid_helper.grid_finder.grid_locator2,MultipleLocator):
                l2 = MultipleLocator(base=yBase)
                ax._grid_helper.update_grid_finder(grid_locator2=l2)
            ax.ybase = yBase

        @updateAxisDecorator(ax=ax,direction='x')
        def set_xticks_number(xNumber,ax=ax):
            """Setter of the number of x ticks to be used for plotting

            Args:

                - xNumber (int): Number of x tick marks

            """
            if not isinstance(ax._grid_helper.grid_finder.grid_locator1,MaxNLocator):
                l1 = MaxNLocator(nbins=xNumber)
                ax._grid_helper.update_grid_finder(grid_locator1=l1)
            ax.xticks = xNumber

        @updateAxisDecorator(ax=ax,direction='y')
        def set_yticks_number(yNumber,ax=ax):
            """Setter of the number of y ticks to be used for plotting

            Args:

                - yNumber (int): Number of y tick marks

            """
            if not isinstance(ax._grid_helper.grid_finder.grid_locator2,MaxNLocator):
                l2 = MaxNLocator(nbins=yNumber)
                ax._grid_helper.update_grid_finder(grid_locator2=l2)
            ax.yticks = yNumber

        ax.set_xticks_base = set_xticks_base
        ax.set_yticks_base = set_yticks_base
        ax.set_xticks_number = set_xticks_number
        ax.set_yticks_number = set_yticks_number

    return ax




def createQEAxes(DataSet=None,axis=0,figure = None, projectionVector1 = None, projectionVector2 = None, rlu = True, figurePos=[1,1,1]):
    """Function to create Q E plot

    Kwargs:

        - DataSet (DataSet): If provided and no projections vectors creates QE axis for main direction (default None)

        - axis (int): Whether to create axis 0 or 1 (projection vector 0 or orthogonal to this, default 0)

        - figure (figure): If provided, this is used to create the axis within (default None)

        - projectionVector1 (vec): Projection vector along which data is plotted. If not provided sample vector is used (default None)

        - projectionVector2 (vec): Projection vector orthogonal to data. If not provided sample vector is used (default None)
        
        - rlu (Bool): Boolean to control if rlu or qxqy (default True)


    """
        
    if projectionVector1 is None or projectionVector2 is None:
        if rlu is True:
            v1 = DataSet.sample[0].projectionVector1
            v2 = DataSet.sample[0].projectionVector2
            angle = DataSet.sample[0].projectionAngle
            orientationMatrix = DataSet.sample[0].orientationMatrix
        else:
            v1 = np.array([1.0,0.0])
            v2 = np.array([0.0,1.0])
            angle = np.pi/2.0
            orientationMatrix = np.ones(2)
    else:
        v1 = np.array(projectionVector1)
        v2 = np.array(projectionVector2)

        if not np.all([x.shape==(3,) for x in [v1,v2]])\
        or not np.all([len(x.shape)==1 for x in [v1,v2]])\
        or np.all([x.shape==(2,) for x in [v1,v2]]):
            raise AttributeError('Provided vector(s) is not 3D/2D: projectionVector1.shape={} or projectionVector2.shape={}'.format(v1.shape,v2.shape))
        angle = np.arccos(np.dot(v1,v2)/(np.linalg.norm(v1)*np.linalg.norm(v2)))
        if rlu is True:
            orientationMatrix = np.ones(3)
        else:
            orientationMatrix = np.ones(2)


    sample = copy.deepcopy(DataSet.sample)


    v2Length = np.linalg.norm(v2)/np.linalg.norm(v1)
    projectionMatrix = np.linalg.inv(np.array([[1,0],[np.cos(angle)*v2Length,np.sin(angle)*v2Length]]).T)

    projectionVectorQX = np.dot(np.dot(projectionMatrix,[1,0]),np.array([v1,v2]))
    projectionVectorQY = np.dot(np.dot(projectionMatrix,[0,v2Length]),np.array([v1,v2]))

    projectionVectorQX = _tools.LengthOrder(projectionVectorQX)
    projectionVectorQY = _tools.LengthOrder(projectionVectorQY)
    projectionVectorQXLength = np.linalg.norm(np.dot(orientationMatrix,projectionVectorQY))
    projectionVectorQYLength = np.linalg.norm(np.dot(orientationMatrix,projectionVectorQX))
    projectionVectorQXFormated = ', '.join(['{:.3f}'.format(x) for x in projectionVectorQX])
    projectionVectorQYFormated = ', '.join(['{:.3f}'.format(x) for x in projectionVectorQY])

    if axis == 0:
        projectionVectorLength = projectionVectorQYLength
        projectionVectorLengthORthogonal = projectionVectorQXLength
        projectionVectorFormated = projectionVectorQXFormated
        projectionVector = projectionVectorQX
        projectionVectorOrthogonal = projectionVectorQY
    elif axis == 1:
        projectionVectorLength = projectionVectorQXLength
        projectionVectorFormated = projectionVectorQYFormated
        projectionVectorLengthORthogonal = projectionVectorQYLength
        projectionVector = projectionVectorQY
        projectionVectorOrthogonal = projectionVectorQX
    else:
        raise AttributeError('Provided axis of {} is not allowed. Should be either 0 or 1.'.format(axis))

    if figure is None:

        figure = plt.figure(figsize=(7, 4))
    #else:
    #    figure.clf()
    def inv_tr(l,x,y):
        return x*l,y

    def tr(l,x,y):
        return x/l,y


    grid_locator1 = MultipleLocator(base=1.0) # Standard X ticks is multiple locator
    grid_helper = GridHelperCurveLinear((lambda x,y:inv_tr(projectionVectorLength,x,y), 
                                    lambda x,y:tr(projectionVectorLength,x,y)),grid_locator1=grid_locator1)

    if rlu:
        labels = ['h','k','l']
    else:
        labels = ['qx','qy']

    ax = SubplotHost(figure, *figurePos, grid_helper=grid_helper)
    
    ax.inv_tr = lambda x,y:inv_tr(projectionVectorLength,x,y)
    ax.tr = lambda x,y:tr(projectionVectorLength,x,y)
    
    ax.sample = sample[0]

    figure.add_subplot(ax)
    #ax.set_aspect(1.)
    ax.grid(True, zorder=0)

    def calculateRLU(l,v1,x,y,v,step):
        return np.asarray(x)/l*v1+v*step, np.asarray(y)

    def format_coord(x,y,rlu=rlu,labels=labels): # pragma: no cover # x is H,K,L and y is  energy
        xformated = ', '.join(['{} = {}'.format(Y[0],Y[1]) for Y in zip(labels,['{:.4f}'.format(X) for X in x])])
        return '{}, E={:.4f}'.format(xformated,y)


    if rlu:
        ax.set_xlabel('{} [RLU]'.format(projectionVectorFormated))
    else:
        ax.set_xlabel('{} [1/A]'.format(projectionVectorFormated))

    ax.set_ylabel('E [meV]')
    ax._length = projectionVectorLengthORthogonal
    ax._projectionVector = projectionVector 
    ax._projectionVectorOrthogonal = projectionVectorOrthogonal
    ax.plotDirection = projectionVector
    ax._step = 0.0
    ax.calculateRLU = lambda x,y: calculateRLU(projectionVectorLength,ax._projectionVector,x,y,ax._projectionVectorOrthogonal,ax._step)
    ax.format_coord = lambda x,y: format_coord(*ax.calculateRLU(x,y))


    ax.forceGridUpdate = lambda:forceGridUpdate(ax)
    ax.xticks = 7

    def xAxisChanged(event, axis=None, forceUpdate=False):
        # Two arguments are needed; event for canvas callback and axis is needed for ax.xAxisChanged
        if axis is None:
            axis = event
        locator = axis._grid_helper.grid_finder.grid_locator1
        xlim = axis.get_xlim()
        xlimDiff = np.diff(xlim)
        if isinstance(locator,MultipleLocator):
            if hasattr(axis,'xBase'):
                base = axis.xBase
            else:
                base = calculateBase(locator,xlimDiff,axis.xticks)
            locator.set_params(base=base)

        elif isinstance(locator,MaxNLocator):
            if hasattr(axis,'xTicks'):
                ticks = getattr(axis,'xTicks')
            else:
                ticks = 7
            locator.set_params(nbins = ticks)
        else:
            return
        axis.forceGridUpdate()

    ax.callbacks.connect('xlim_changed', xAxisChanged)

    ax.get_figure().canvas.mpl_connect('draw_event',lambda event: xAxisChanged(event,axis=ax,forceUpdate=True))
    ax.xAxisChanged = lambda: xAxisChanged(None,ax,forceUpdate=True)


    @updateXAxisDecorator(ax=ax)
    def set_xticks_base(xBase=None,ax=ax):
        """Setter of the base x ticks to be used for plotting

        Kwargs:

            - xBase (float): Base of the tick marks (default automatic)

        """


        if not isinstance(ax._grid_helper.grid_finder.grid_locator1,MultipleLocator):
            l1 = MultipleLocator(base=xBase)
            ax._grid_helper.update_grid_finder(grid_locator1=l1)

        if xBase is None:
            if hasattr(ax,'xBase'):
                delattr(ax,'xBase')
        else:
            ax.xBase = xBase

    @updateXAxisDecorator(ax=ax)
    def set_xticks_number(xNumber = None,ax=ax):
        """Setter of the number of x ticks to be used for plotting

        Kwargs:

            - xNumber (int): Number of x tick marks (default 7)

        """
        if xNumber is None:
            xNumber = 7

        if not isinstance(ax._grid_helper.grid_finder.grid_locator1,MaxNLocator):
            l1 = MaxNLocator(nbins=xNumber)
            ax._grid_helper.update_grid_finder(grid_locator1=l1)
        ax.xTicks = xNumber


    ax.set_xticks_base = set_xticks_base
    ax.set_xticks_number = set_xticks_number

    return ax









#!/usr/bin/env python
#
# resolution ellipsoid calculations
# 
# @author Tobias Weber <tweber@ill.fr>
# @date mar-2019
# @license GPLv2
#
# @desc for algorithm: [eck14] G. Eckold and O. Sobolev, NIM A 752, pp. 54-64 (2014)
# @desc see covariance calculations: https://code.ill.fr/scientific-software/takin/mag-core/blob/master/tools/tascalc/cov.py
#

# requires numpy version >= 1.10
import numpy as np
import numpy.linalg as la
import os

#np.set_printoptions(floatmode = "fixed",  precision = 4)


sig2fwhm = 2.*np.sqrt(2.*np.log(2.))


#
# volume of the ellipsoid
#
def ellipsoid_volume(mat):
    det = np.abs(la.det(mat))

    return 4./3. * np.pi * np.sqrt(1./det)


#
# projects along one axis of the quadric
# (see [eck14], equ. 57)
#
def quadric_proj(_E, idx):
    E = np.delete(np.delete(_E, idx, axis=0), idx, axis=1)
    if np.abs(_E[idx, idx]) < 1e-8:
        return E

    v = 0.5 * (_E[idx,:] + _E[:,idx])
    vv = np.outer(v, v) / _E[idx, idx]
    vv = np.delete(np.delete(vv, idx, axis=0), idx, axis=1)

    return E - vv


#
# projects linear part of the quadric
# (see [eck14], equ. 57)
#
def quadric_proj_vec(vec, _E, idx):
    _col = _E[:,idx]
    col = np.delete(_col, idx, axis=0)
    if np.abs(_col[idx]) < 1e-8:
        return col

    v = np.delete(vec, idx, axis=0)
    v = v - col*vec[idx]/_col[idx]

    return v


#
# coherent fwhm widths
#
def calc_coh_fwhms(reso):
    vecFwhms = []
    for i in range(len(reso)):
        vecFwhms.append(sig2fwhm / np.sqrt(reso[i,i]))

    return np.array(vecFwhms)


#
# incoherent fwhm width
#
def calc_incoh_fwhm(reso):
	Qres_proj = quadric_proj(reso, 2)
	Qres_proj = quadric_proj(Qres_proj, 1)
	Qres_proj = quadric_proj(Qres_proj, 0)

	return 1./np.sqrt(np.abs(Qres_proj[0,0])) * sig2fwhm


#
# calculates the characteristics of a given ellipse by principal axis trafo
#
def descr_ellipse(quadric):
	[ evals, evecs ] = la.eig(quadric)

	fwhms = 1./np.sqrt(np.abs(evals)) * sig2fwhm

	angles = np.array([])
	if len(quadric) == 2:
		angles = np.array([ np.arctan2(evecs[1][0], evecs[0][0]) ])

	return [fwhms, angles/np.pi*180., evecs]


#
# describes the ellipsoid by a principal axis trafo and by 2d cuts
#
def calc_ellipses(Qres_Q, verbose=True):
	# 4d ellipsoid
	[fwhms, angles, rot] = descr_ellipse(Qres_Q)

	if verbose:
		print()
		print("Coherent-elastic fwhms: %s" % (calc_coh_fwhms(Qres_Q)))
		print("Incoherent-elastic fwhm: %.4f meV" % (calc_incoh_fwhm(Qres_Q)))
		print("Principal axes fwhms: %s" % fwhms)


	# 2d sliced ellipses
	if verbose:
		print()
	Qres_QxE = np.delete(np.delete(Qres_Q, 2, axis=0), 2, axis=1)
	Qres_QxE = np.delete(np.delete(Qres_QxE, 1, axis=0), 1, axis=1)
	[fwhms_QxE, angles_QxE, rot_QxE] = descr_ellipse(Qres_QxE)
	if verbose:
		print("2d Qx,E slice fwhms and slope angle: %s, %.4f" % (fwhms_QxE, angles_QxE[0]))

	Qres_QyE = np.delete(np.delete(Qres_Q, 2, axis=0), 2, axis=1)
	Qres_QyE = np.delete(np.delete(Qres_QyE, 0, axis=0), 0, axis=1)
	[fwhms_QyE, angles_QyE, rot_QyE] = descr_ellipse(Qres_QyE)
	if verbose:
		print("2d Qy,E slice fwhms and slope angle: %s, %.4f" % (fwhms_QyE, angles_QyE[0]))

	Qres_QzE = np.delete(np.delete(Qres_Q, 1, axis=0), 1, axis=1)
	Qres_QzE = np.delete(np.delete(Qres_QzE, 0, axis=0), 0, axis=1)
	[fwhms_QzE, angles_QzE, rot_QzE] = descr_ellipse(Qres_QzE)
	if verbose:
		print("2d Qz,E slice fwhms and slope angle: %s, %.4f" % (fwhms_QzE, angles_QzE[0]))

	Qres_QxQy = np.delete(np.delete(Qres_Q, 3, axis=0), 3, axis=1)
	Qres_QxQy = np.delete(np.delete(Qres_QxQy, 2, axis=0), 2, axis=1)
	[fwhms_QxQy, angles_QxQy, rot_QxQy] = descr_ellipse(Qres_QxQy)
	if verbose:
		print("2d Qx,Qy slice fwhms and slope angle: %s, %.4f" % (fwhms_QxQy, angles_QxQy[0]))


	# 2d projected ellipses
	if verbose:
		print()
	Qres_QxE_proj = np.delete(np.delete(Qres_Q, 2, axis=0), 2, axis=1)
	Qres_QxE_proj = quadric_proj(Qres_QxE_proj, 1)
	[fwhms_QxE_proj, angles_QxE_proj, rot_QxE_proj] = descr_ellipse(Qres_QxE_proj)
	if verbose:
		print("2d Qx,E projection fwhms and slope angle: %s, %.4f" % (fwhms_QxE_proj, angles_QxE_proj[0]))

	Qres_QyE_proj = np.delete(np.delete(Qres_Q, 2, axis=0), 2, axis=1)
	Qres_QyE_proj = quadric_proj(Qres_QyE_proj, 0)
	[fwhms_QyE_proj, angles_QyE_proj, rot_QyE_proj] = descr_ellipse(Qres_QyE_proj)
	if verbose:
		print("2d Qy,E projection fwhms and slope angle: %s, %.4f" % (fwhms_QyE_proj, angles_QyE_proj[0]))

	Qres_QzE_proj = np.delete(np.delete(Qres_Q, 1, axis=0), 1, axis=1)
	Qres_QzE_proj = quadric_proj(Qres_QzE_proj, 0)
	[fwhms_QzE_proj, angles_QzE_proj, rot_QzE_proj] = descr_ellipse(Qres_QzE_proj)
	if verbose:
		print("2d Qz,E projection fwhms and slope angle: %s, %.4f" % (fwhms_QzE_proj, angles_QzE_proj[0]))

	Qres_QxQy_proj = quadric_proj(Qres_Q, 3)
	Qres_QxQy_proj = np.delete(np.delete(Qres_QxQy_proj, 2, axis=0), 2, axis=1)
	[fwhms_QxQy_proj, angles_QxQy_proj, rot_QxQy_proj] = descr_ellipse(Qres_QxQy_proj)
	if verbose:
		print("2d Qx,Qy projection fwhms and slope angle: %s, %.4f" % (fwhms_QxQy_proj, angles_QxQy_proj[0]))


	results = {
		"fwhms_QxE" : fwhms_QxE, "rot_QxE" : rot_QxE, 
		"fwhms_QyE" : fwhms_QyE, "rot_QyE" : rot_QyE,
		"fwhms_QzE" : fwhms_QzE, "rot_QzE" : rot_QzE, 
		"fwhms_QxQy" : fwhms_QxQy,  "rot_QxQy" : rot_QxQy,
		"fwhms_QxE_proj" : fwhms_QxE_proj,  "rot_QxE_proj" : rot_QxE_proj, 
		"fwhms_QyE_proj" : fwhms_QyE_proj, "rot_QyE_proj" : rot_QyE_proj,
		"fwhms_QzE_proj" : fwhms_QzE_proj, "rot_QzE_proj" : rot_QzE_proj, 
		"fwhms_QxQy_proj" : fwhms_QxQy_proj, "rot_QxQy_proj" : rot_QxQy_proj,
	}

	return results



#
# shows the 2d ellipses
#
def plot_ellipses(ellis, verbose=True, plot_results=True, file="", dpi=600, ellipse_points=128):
	import mpl_toolkits.mplot3d as mplot3d
	import matplotlib.pyplot as plot

	ellfkt = lambda rad, rot, phi : \
		np.dot(rot, np.array([ rad[0]*np.cos(phi), rad[1]*np.sin(phi) ]))


	phi = np.linspace(0, 2.*np.pi, ellipse_points)

	ell_QxE = ellfkt(ellis["fwhms_QxE"]*0.5, ellis["rot_QxE"], phi)
	ell_QyE = ellfkt(ellis["fwhms_QyE"]*0.5, ellis["rot_QyE"], phi)
	ell_QzE = ellfkt(ellis["fwhms_QzE"]*0.5, ellis["rot_QzE"], phi)
	ell_QxQy = ellfkt(ellis["fwhms_QxQy"]*0.5, ellis["rot_QxQy"], phi)

	ell_QxE_proj = ellfkt(ellis["fwhms_QxE_proj"]*0.5, ellis["rot_QxE_proj"], phi)
	ell_QyE_proj = ellfkt(ellis["fwhms_QyE_proj"]*0.5, ellis["rot_QyE_proj"], phi)
	ell_QzE_proj = ellfkt(ellis["fwhms_QzE_proj"]*0.5, ellis["rot_QzE_proj"], phi)
	ell_QxQy_proj = ellfkt(ellis["fwhms_QxQy_proj"]*0.5, ellis["rot_QxQy_proj"], phi)


	# Qpara, E axis
	fig = plot.figure()
	subplot_QxE = fig.add_subplot(221)
	subplot_QxE.set_xlabel("Qpara (1/A)")
	subplot_QxE.set_ylabel("E (meV)")
	subplot_QxE.plot(ell_QxE[0], ell_QxE[1], c="black", linestyle="dashed")
	subplot_QxE.plot(ell_QxE_proj[0], ell_QxE_proj[1], c="black", linestyle="solid")

	# Qperp, E axis
	subplot_QyE = fig.add_subplot(222)
	subplot_QyE.set_xlabel("Qperp (1/A)")
	subplot_QyE.set_ylabel("E (meV)")
	subplot_QyE.plot(ell_QyE[0], ell_QyE[1], c="black", linestyle="dashed")
	subplot_QyE.plot(ell_QyE_proj[0], ell_QyE_proj[1], c="black", linestyle="solid")

	# Qup, E axis
	subplot_QzE = fig.add_subplot(223)
	subplot_QzE.set_xlabel("Qup (1/A)")
	subplot_QzE.set_ylabel("E (meV)")
	subplot_QzE.plot(ell_QzE[0], ell_QzE[1], c="black", linestyle="dashed")
	subplot_QzE.plot(ell_QzE_proj[0], ell_QzE_proj[1], c="black", linestyle="solid")

	# Qpara, Qperp axis
	subplot_QxQy = fig.add_subplot(224)
	subplot_QxQy.set_xlabel("Qpara (1/A)")
	subplot_QxQy.set_ylabel("Qperp (1/A)")
	subplot_QxQy.plot(ell_QxQy[0], ell_QxQy[1], c="black", linestyle="dashed")
	subplot_QxQy.plot(ell_QxQy_proj[0], ell_QxQy_proj[1], c="black", linestyle="solid")
	plot.tight_layout()


	# 3d plot
	fig3d = plot.figure()
	subplot3d = fig3d.add_subplot(111, projection="3d")

	subplot3d.set_xlabel("Qpara (1/A)")
	subplot3d.set_ylabel("Qperp (1/A)")
	subplot3d.set_zlabel("E (meV)")

	# xE
	subplot3d.plot(ell_QxE[0], ell_QxE[1], zs=0., zdir="y", c="black", linestyle="dashed")
	subplot3d.plot(ell_QxE_proj[0], ell_QxE_proj[1], zs=0., zdir="y", c="black", linestyle="solid")
	# yE
	subplot3d.plot(ell_QyE[0], ell_QyE[1], zs=0., zdir="x", c="black", linestyle="dashed")
	subplot3d.plot(ell_QyE_proj[0], ell_QyE_proj[1], zs=0., zdir="x", c="black", linestyle="solid")
	# xy
	subplot3d.plot(ell_QxQy[0], ell_QxQy[1], zs=0., zdir="z", c="black", linestyle="dashed")
	subplot3d.plot(ell_QxQy_proj[0], ell_QxQy_proj[1], zs=0., zdir="z", c="black", linestyle="solid")


	# save plots to files
	if file != "":
		splitext = os.path.splitext(file)
		file3d = splitext[0] + "_3d" + splitext[1]

		if verbose:
			print("Saving 2d plot to \"%s\"." % file)
			print("Saving 3d plot to \"%s\"." % file3d)
		fig.savefig(file, dpi=dpi)
		fig3d.savefig(file3d, dpi=dpi)


	# show plots
	if plot_results:
		plot.show()

	return fig,fig3d